# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click

from bentoml.utils import status_pb_to_error_code_and_message
from bentoml.utils.lazy_loader import LazyLoader
from bentoml.utils import get_default_yatai_client
from bentoml.cli.utils import Spinner
from bentoml.cli.click_utils import (
    BentoMLCommandGroup,
    parse_bento_tag_callback,
    _echo,
    CLI_COLOR_SUCCESS,
    parse_labels_callback,
)
from bentoml.cli.deployment import (
    _print_deployment_info,
    _print_deployments_info,
)
from bentoml.yatai.deployment import ALL_NAMESPACE_TAG
from bentoml.exceptions import CLIException

yatai_proto = LazyLoader('yatai_proto', globals(), 'bentoml.yatai.proto')


def get_aws_lambda_sub_command():
    # pylint: disable=unused-variable

    @click.group(
        name='lambda',
        help='Commands for AWS Lambda BentoService deployments',
        cls=BentoMLCommandGroup,
    )
    def aws_lambda():
        pass

    @aws_lambda.command(help='Deploy BentoService to AWS Lambda')
    @click.argument('name', type=click.STRING)
    @click.option(
        '-b',
        '--bento',
        '--bento-service-bundle',
        type=click.STRING,
        required=True,
        callback=parse_bento_tag_callback,
        help='Target BentoService to be deployed, referenced by its name and version '
        'in format of name:version. For example: "iris_classifier:v1.2.0"',
    )
    @click.option(
        '-n',
        '--namespace',
        type=click.STRING,
        help='Deployment namespace managed by BentoML, default value is "dev" which '
        'can be changed in BentoML configuration yatai_service/default_namespace',
    )
    @click.option(
        '-l',
        '--labels',
        type=click.STRING,
        callback=parse_labels_callback,
        help='Key:value pairs that are attached to deployments and intended to be used '
        'to specify identifying attributes of the deployments that are meaningful to '
        'users. Multiple labels are separated with `,`',
    )
    @click.option('--region', help='AWS region name for deployment')
    @click.option(
        '--api-name', help='User defined API function will be used for inference',
    )
    @click.option(
        '--memory-size',
        help="Maximum Memory Capacity for AWS Lambda function, you can set the memory "
        "size in 64MB increments from 128MB to 3008MB. The default value "
        "is 1024 MB.",
        type=click.INT,
        default=1024,
    )
    @click.option(
        '--timeout',
        help="The amount of time that AWS Lambda allows a function to run before "
        "stopping it. The default is 3 seconds. The maximum allowed value is "
        "900 seconds",
        type=click.INT,
        default=3,
    )
    @click.option('-o', '--output', type=click.Choice(['json', 'yaml']), default='json')
    @click.option(
        '--wait/--no-wait',
        default=True,
        help='Wait for apply action to complete or encounter an error.'
        'If set to no-wait, CLI will return immediately. The default value is wait',
    )
    def deploy(
        name,
        bento,
        namespace,
        labels,
        region,
        api_name,
        memory_size,
        timeout,
        output,
        wait,
    ):
        _echo(
            message='AWS Lambda deployment functionalities are being migrated to a '
            'separate tool and related CLI commands will be deprecated in BentoML '
            'itself, please use https://github.com/bentoml/aws-lambda-deploy '
            'going forward.',
            color='yellow',
        )
        yatai_client = get_default_yatai_client()
        bento_name, bento_version = bento.split(':')
        with Spinner(f'Deploying "{bento}" to AWS Lambda '):
            result = yatai_client.deployment.create_lambda_deployment(
                name=name,
                namespace=namespace,
                bento_name=bento_name,
                bento_version=bento_version,
                api_name=api_name,
                region=region,
                memory_size=memory_size,
                timeout=timeout,
                labels=labels,
                wait=wait,
            )
        if result.status.status_code != yatai_proto.status_pb2.Status.OK:
            error_code, error_message = status_pb_to_error_code_and_message(
                result.status
            )
            raise CLIException(f'{error_code}:{error_message}')
        _echo(f'Successfully created AWS Lambda deployment {name}', CLI_COLOR_SUCCESS)
        _print_deployment_info(result.deployment, output)

    @aws_lambda.command(help='Update existing AWS Lambda deployment')
    @click.argument('name', type=click.STRING)
    @click.option(
        '-b',
        '--bento',
        '--bento-service-bundle',
        type=click.STRING,
        callback=parse_bento_tag_callback,
        help='Target BentoService to be deployed, referenced by its name and version '
        'in format of name:version. For example: "iris_classifier:v1.2.0"',
    )
    @click.option(
        '-n',
        '--namespace',
        type=click.STRING,
        help='Deployment namespace managed by BentoML, default value is "dev" which '
        'can be changed in BentoML configuration yatai_service/default_namespace',
    )
    @click.option(
        '--memory-size',
        help="Maximum memory capacity for AWS Lambda function in MB, you can set "
        "the memory size in 64MB increments from 128 to 3008. "
        "The default value is 1024",
        type=click.INT,
    )
    @click.option(
        '--timeout',
        help="The amount of time that AWS Lambda allows a function to run before "
        "stopping it. The default is 3 seconds. The maximum allowed value is "
        "900 seconds",
        type=click.INT,
    )
    @click.option('-o', '--output', type=click.Choice(['json', 'yaml']), default='json')
    @click.option(
        '--wait/--no-wait',
        default=True,
        help='Wait for apply action to complete or encounter an error.'
        'If set to no-wait, CLI will return immediately. The default value is wait',
    )
    def update(name, namespace, bento, memory_size, timeout, output, wait):
        _echo(
            message='AWS Lambda deployment functionalities are being migrated to a '
            'separate tool and related CLI commands will be deprecated in BentoML '
            'itself, please use https://github.com/bentoml/aws-lambda-deploy '
            'going forward.',
            color='yellow',
        )
        yatai_client = get_default_yatai_client()
        if bento:
            bento_name, bento_version = bento.split(':')
        else:
            bento_name = None
            bento_version = None
        with Spinner('Updating Lambda deployment '):
            result = yatai_client.deployment.update_lambda_deployment(
                bento_name=bento_name,
                bento_version=bento_version,
                deployment_name=name,
                namespace=namespace,
                memory_size=memory_size,
                timeout=timeout,
                wait=wait,
            )
        if result.status.status_code != yatai_proto.status_pb2.Status.OK:
            error_code, error_message = status_pb_to_error_code_and_message(
                result.status
            )
            raise CLIException(f'{error_code}:{error_message}')
        _echo(
            f'Successfully updated AWS Lambda deployment {name}', CLI_COLOR_SUCCESS,
        )
        _print_deployment_info(result.deployment, output)

    @aws_lambda.command(help='Delete AWS Lambda deployment')
    @click.argument('name', type=click.STRING)
    @click.option(
        '-n',
        '--namespace',
        type=click.STRING,
        help='Deployment namespace managed by BentoML, default value is "dev" which '
        'can be changed in BentoML configuration yatai_service/default_namespace',
    )
    @click.option(
        '--force',
        is_flag=True,
        help='force delete the deployment record in database and '
        'ignore errors when deleting cloud resources',
    )
    def delete(name, namespace, force):
        _echo(
            message='AWS Lambda deployment functionalities are being migrated to a '
            'separate tool and related CLI commands will be deprecated in BentoML '
            'itself, please use https://github.com/bentoml/aws-lambda-deploy '
            'going forward.',
            color='yellow',
        )
        yatai_client = get_default_yatai_client()
        get_deployment_result = yatai_client.deployment.get(
            namespace=namespace, name=name
        )
        if get_deployment_result.status.status_code != yatai_proto.status_pb2.Status.OK:
            error_code, error_message = status_pb_to_error_code_and_message(
                get_deployment_result.status
            )
            raise CLIException(f'{error_code}:{error_message}')
        result = yatai_client.deployment.delete(
            namespace=namespace, deployment_name=name, force_delete=force
        )
        if result.status.status_code != yatai_proto.status_pb2.Status.OK:
            error_code, error_message = status_pb_to_error_code_and_message(
                result.status
            )
            raise CLIException(f'{error_code}:{error_message}')
        _echo(
            f'Successfully deleted AWS Lambda deployment "{name}"', CLI_COLOR_SUCCESS,
        )

    @aws_lambda.command(help='Get AWS Lambda deployment information')
    @click.argument('name', type=click.STRING)
    @click.option(
        '-n',
        '--namespace',
        type=click.STRING,
        help='Deployment namespace managed by BentoML, default value is "dev" which '
        'can be changed in BentoML configuration yatai_service/default_namespace',
    )
    @click.option(
        '-o', '--output', type=click.Choice(['json', 'yaml', 'table']), default='json'
    )
    def get(name, namespace, output):
        _echo(
            message='AWS Lambda deployment functionalities are being migrated to a '
            'separate tool and related CLI commands will be deprecated in BentoML '
            'itself, please use https://github.com/bentoml/aws-lambda-deploy '
            'going forward.',
            color='yellow',
        )
        yatai_client = get_default_yatai_client()
        describe_result = yatai_client.deployment.describe(namespace, name)
        if describe_result.status.status_code != yatai_proto.status_pb2.Status.OK:
            error_code, error_message = status_pb_to_error_code_and_message(
                describe_result.status
            )
            raise CLIException(f'{error_code}:{error_message}')

        get_result = yatai_client.deployment.get(namespace, name)
        if get_result.status.status_code != yatai_proto.status_pb2.Status.OK:
            error_code, error_message = status_pb_to_error_code_and_message(
                get_result.status
            )
            raise CLIException(f'{error_code}:{error_message}')
        _print_deployment_info(get_result.deployment, output)

    @aws_lambda.command(name='list', help='List AWS Lambda deployments')
    @click.option(
        '-n',
        '--namespace',
        type=click.STRING,
        help='Deployment namespace managed by BentoML, default value is "dev" which '
        'can be changed in BentoML configuration yatai_service/default_namespace',
        default=ALL_NAMESPACE_TAG,
    )
    @click.option(
        '--limit',
        type=click.INT,
        help='The maximum amount of AWS Lambda deployments to be listed at once',
    )
    @click.option(
        '-l',
        '--labels',
        type=click.STRING,
        help="Label query to filter Lambda deployments, supports '=', '!=', 'IN', "
        "'NotIn', 'Exists', and 'DoesNotExist'. (e.g. key1=value1, "
        "key2!=value2, key3 In (value3, value3a), key4 DoesNotExist)",
    )
    @click.option(
        '--order-by', type=click.Choice(['created_at', 'name']), default='created_at',
    )
    @click.option(
        '--asc/--desc',
        default=False,
        help='Ascending or descending order for list deployments',
    )
    @click.option(
        '-o',
        '--output',
        type=click.Choice(['json', 'yaml', 'table', 'wide']),
        default='table',
    )
    def list_deployments(namespace, limit, labels, order_by, asc, output):
        _echo(
            message='AWS Lambda deployment functionalities are being migrated to a '
            'separate tool and related CLI commands will be deprecated in BentoML '
            'itself, please use https://github.com/bentoml/aws-lambda-deploy '
            'going forward.',
            color='yellow',
        )
        yatai_client = get_default_yatai_client()
        list_result = yatai_client.deployment.list_lambda_deployments(
            limit=limit,
            labels=labels,
            namespace=namespace,
            order_by=order_by,
            ascending_order=asc,
        )
        if list_result.status.status_code != yatai_proto.status_pb2.Status.OK:
            error_code, error_message = status_pb_to_error_code_and_message(
                list_result.status
            )
            raise CLIException(f'{error_code}:{error_message}')
        else:
            _print_deployments_info(list_result.deployments, output)

    return aws_lambda



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Tuple, Type

from bentoml.types import (
    ApiFuncReturnValue,
    AwsLambdaEvent,
    HTTPResponse,
    InferenceResult,
    InferenceTask,
)

from .base_output import BaseOutputAdapter


def detect_suitable_adapter(result) -> Type[BaseOutputAdapter]:
    try:
        import pandas as pd

        if isinstance(result, (pd.DataFrame, pd.Series)):
            from .dataframe_output import DataframeOutput

            return DataframeOutput
    except ImportError:
        pass

    try:
        import tensorflow as tf

        if isinstance(result, tf.Tensor):
            from .tensorflow_tensor_output import TfTensorOutput

            return TfTensorOutput
    except ImportError:
        pass

    from .json_output import JsonOutput

    return JsonOutput


class DefaultOutput(BaseOutputAdapter):
    """
    Detect suitable output adapter automatically
    OutputAdapter converts returns of user defined API function into specific output,
    such as HTTP response, command line stdout or AWS Lambda event object.

    Args:
        cors (str): DEPRECATED. Moved to the configuration file.
            The value of the Access-Control-Allow-Origin header set in the
            HTTP/AWS Lambda response object. If set to None, the header will not be set.
            Default is None.
    """

    def __init__(self, **kwargs):
        super(DefaultOutput, self).__init__(**kwargs)
        self.actual_adapter = None
        from .json_output import JsonOutput

        self.backup_adapter = JsonOutput()

    def pack_user_func_return_value(
        self, return_result: ApiFuncReturnValue, tasks: Tuple[InferenceTask],
    ) -> Tuple[InferenceResult]:
        if self.actual_adapter is None:
            self.actual_adapter = detect_suitable_adapter(return_result)()
        if self.actual_adapter:
            return self.actual_adapter.pack_user_func_return_value(
                return_result, tasks=tasks
            )

        raise NotImplementedError()

    def to_http_response(self, result) -> HTTPResponse:
        if self.actual_adapter:
            return self.actual_adapter.to_http_response(result)
        return self.backup_adapter.to_http_response(result)

    def to_cli(self, results) -> int:
        if self.actual_adapter:
            return self.actual_adapter.to_cli(results)
        return self.backup_adapter.to_cli(results)

    def to_aws_lambda_event(self, result) -> AwsLambdaEvent:
        if self.actual_adapter:
            return self.actual_adapter.to_aws_lambda_event(result)
        return self.backup_adapter.to_aws_lambda_event(result)



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from bentoml.yatai.repository.base_repository import BaseRepository
from bentoml.yatai.repository.file_system_repository import FileSystemRepository
from bentoml.yatai.repository.s3_repository import S3Repository
from bentoml.yatai.repository.gcs_repository import GCSRepository


def create_repository(
    repository_type: str,
    file_system_directory=None,
    s3_url=None,
    s3_endpoint_url=None,
    gcs_url=None,
) -> BaseRepository:
    """Creates a repository based on a provided type and parameters"""
    if repository_type == "s3":
        return S3Repository(s3_url, endpoint_url=s3_endpoint_url)
    elif repository_type == "gcs":
        return GCSRepository(gcs_url)
    elif repository_type == "file_system":
        return FileSystemRepository(file_system_directory)
    else:
        raise ValueError("Unrecognized repository type {}" % repository_type)



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cerberus import Validator

from bentoml.exceptions import InvalidArgument
from bentoml.utils import ProtoMessageToDict
from bentoml.yatai.db.stores.label import _validate_labels
from bentoml.yatai.deployment.azure_functions.constants import (
    AZURE_FUNCTIONS_PREMIUM_PLAN_SKUS,
    AZURE_FUNCTIONS_AUTH_LEVELS,
)
from bentoml.yatai.proto.deployment_pb2 import DeploymentSpec, DeploymentState

deployment_schema = {
    'name': {'type': 'string', 'required': True, 'minlength': 4},
    # namespace is optional - YataiService will fill-in the default namespace configured
    # when it is missing in the apply deployment request
    'namespace': {'type': 'string', 'required': False, 'minlength': 3},
    'labels': {'type': 'dict', 'deployment_labels': True},
    'annotations': {'type': 'dict', 'allow_unknown': True},
    'created_at': {'type': 'string'},
    'last_updated_at': {'type': 'string'},
    'spec': {
        'type': 'dict',
        'required': True,
        'schema': {
            'operator': {
                'type': 'string',
                'required': True,
                'allowed': DeploymentSpec.DeploymentOperator.keys(),
            },
            'bento_name': {'type': 'string', 'required': True},
            'bento_version': {
                'type': 'string',
                'required': True,
                'bento_service_version': True,
            },
            'custom_operator_config': {
                'type': 'dict',
                'schema': {
                    'name': {'type': 'string'},
                    'config': {'type': 'dict', 'allow_unknown': True},
                },
            },
            'sagemaker_operator_config': {
                'type': 'dict',
                'schema': {
                    'api_name': {'type': 'string', 'required': True, 'minlength': 3},
                    'instance_type': {'type': 'string', 'required': True},
                    'instance_count': {'type': 'integer', 'min': 1, 'required': True},
                    'region': {'type': 'string'},
                    'num_of_gunicorn_workers_per_instance': {
                        'type': 'integer',
                        'min': 1,
                    },
                    'timeout': {'type': 'integer', 'min': 1},
                    'data_capture_s3_prefix': {'type': 'string'},
                    'data_capture_sample_percent': {
                        'type': 'integer',
                        'min': 1,
                        'max': 100,
                    },
                },
            },
            'aws_lambda_operator_config': {
                'type': 'dict',
                'schema': {
                    'region': {'type': 'string'},
                    'api_name': {'type': 'string', 'minlength': 3},
                    'memory_size': {'type': 'integer', 'aws_lambda_memory': True},
                    'timeout': {'type': 'integer', 'min': 1, 'max': 900},
                },
            },
            # https://docs.microsoft.com/en-us/azure/azure-functions/functions-premium-plan
            # https://docs.microsoft.com/en-us/azure/azure-functions/functions-bindings-http-webhook-trigger?tabs=python#configuration
            'azure_functions_operator_config': {
                'type': 'dict',
                'azure_functions_configuration': True,
                'schema': {
                    'location': {'type': 'string'},
                    'premium_plan_sku': {
                        'type': 'string',
                        'allowed': AZURE_FUNCTIONS_PREMIUM_PLAN_SKUS,
                    },
                    'min_instances': {
                        'required': True,
                        'type': 'integer',
                        'min': 1,
                        'max': 20,
                    },
                    'max_burst': {
                        'type': 'integer',
                        'min': 1,
                        'max': 20,
                        'required': True,
                    },
                    'function_auth_level': {
                        'type': 'string',
                        'allowed': AZURE_FUNCTIONS_AUTH_LEVELS,
                    },
                },
            },
            "aws_ec2_operator_config": {
                "type": "dict",
                "aws_ec2_operator_configurations": True,
                "schema": {
                    "region": {"type": "string"},
                    "instance_type": {"type": "string"},
                    "ami_id": {"type": "string"},
                    "autoscale_min_size": {
                        "type": "integer",
                        "min": 0,
                        "required": True,
                    },
                    "autoscale_desired_capacity": {
                        "type": "integer",
                        "min": 0,
                        "required": True,
                    },
                    "autoscale_max_size": {
                        "type": "integer",
                        "min": 0,
                        "required": True,
                    },
                },
            },
        },
    },
    'state': {
        'type': 'dict',
        'schema': {
            'state': {'type': 'string', 'allowed': DeploymentState.State.keys()},
            'error_message': {'type': 'string'},
            'info_json': {'type': 'string'},
            'timestamp': {'type': 'string'},
        },
    },
}


class YataiDeploymentValidator(Validator):
    def _validate_aws_lambda_memory(self, aws_lambda_memory, field, value):
        """ Test the memory size restriction for AWS Lambda.

        The rule's arguments are validated against this schema:
        {'type': 'boolean'}
        """
        if aws_lambda_memory:
            if value > 3008 or value < 128:
                self._error(
                    field,
                    'AWS Lambda memory must be between 128 MB to 3,008 MB, '
                    'in 64 MB increments.',
                )
            if value % 64 > 0:
                self._error(
                    field,
                    'AWS Lambda memory must be between 128 MB to 3,008 MB, '
                    'in 64 MB increments.',
                )

    def _validate_bento_service_version(self, bento_service_version, field, value):
        """ Test the given BentoService version is not "latest"

        The rule's arguments are validated against this schema:
        {'type': 'boolean'}
        """
        if bento_service_version and value.lower() == "latest":
            self._error(
                field,
                'Must use specific "bento_version" in deployment, using "latest" is '
                'an anti-pattern.',
            )

    def _validate_azure_functions_configuration(
        self, azure_functions_configuration, field, value
    ):
        """ Test the min instances is less than max burst for Azure Functions deployment

        The rule's arguments are validated against this schema:
        {'type': 'boolean'}
        """
        if azure_functions_configuration:
            if value.get('max_burst', 0) < value.get('min_instances', 0):
                self._error(
                    field,
                    'Azure Functions min instances must be smaller than max burst',
                )

    def _validate_deployment_labels(self, deployment_labels, field, value):
        """ Test label key value schema

        The rule's arguments are validated against this schema:
        {'type': 'boolean'}
        """
        if deployment_labels:
            try:
                _validate_labels(value)
            except InvalidArgument:
                self._error(
                    field,
                    'Valid label key and value must be 63 characters or less and '
                    'must be being and end with an alphanumeric character '
                    '[a-z0-9A-Z] with dashes (-), underscores (_), and dots (.)',
                )

    def _validate_aws_ec2_operator_configurations(
        self, aws_ec2_operator_configurations, field, value
    ):
        """ Test label key value schema

        The rule's arguments are validated against this schema:
        {'type': 'boolean'}
        """
        if aws_ec2_operator_configurations:
            if (
                value.get("autoscale_min_size") < 0
                or value.get("autoscale_max_size") < value.get("autoscale_min_size")
                or value.get("autoscale_desired_capacity")
                < value.get("autoscale_min_size")
                or value.get("autoscale_desired_capacity")
                > value.get("autoscale_max_size")
            ):
                self._error(
                    field,
                    "Wrong autoscaling size specified. "
                    "It should be min_size <= desired_capacity <= max_size",
                )


def validate_deployment_pb(pb):
    pb_dict = ProtoMessageToDict(pb)
    v = YataiDeploymentValidator(deployment_schema)
    if v.validate(pb_dict):
        return None
    else:
        return v.errors



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import logging
import pkgutil
import ast
import zipimport
import zipfile
from typing import Dict

EPP_NO_ERROR = 0
EPP_PKG_NOT_EXIST = 1
EPP_PKG_VERSION_MISMATCH = 2

ZIPIMPORT_DIR = 'zipimports'

__mm = None


logger = logging.getLogger(__name__)


def parse_requirement_string(rs):
    name, _, version = rs.partition("==")
    return name, version


def verify_pkg(pkg_req):
    global __mm  # pylint: disable=global-statement
    if __mm is None:
        __mm = ModuleManager()
    return __mm.verify_pkg(pkg_req)


def seek_pip_packages(target_py_file_path):
    global __mm  # pylint: disable=global-statement
    if __mm is None:
        __mm = ModuleManager()
    return __mm.seek_pip_packages(target_py_file_path)


def get_pkg_version(pkg_name):
    global __mm  # pylint: disable=global-statement
    if __mm is None:
        __mm = ModuleManager()
    return __mm.pip_pkg_map.get(pkg_name, None)


def get_zipmodules():
    global __mm  # pylint: disable=global-statement
    if __mm is None:
        __mm = ModuleManager()
    return __mm.zip_modules


def get_all_pip_installed_modules():
    global __mm  # pylint: disable=global-statement
    if __mm is None:
        __mm = ModuleManager()

    installed_modules = list(
        # local modules are the ones imported from current directory, either from a
        # module.py file or a module directory that contains a `__init__.py` file
        filter(lambda m: not m.is_local, __mm.searched_modules.values())
    )
    return list(map(lambda m: m.name, installed_modules))


class ModuleInfo(object):
    def __init__(self, name, path, is_local, is_pkg):
        super(ModuleInfo, self).__init__()
        self.name = name
        self.path = path
        self.is_local = is_local
        self.is_pkg = is_pkg


class ModuleManager(object):
    def __init__(self):
        super(ModuleManager, self).__init__()
        self.pip_pkg_map = {}
        self.pip_module_map = {}
        self.setuptools_module_set = set()
        self.nonlocal_package_path = set()

        import pkg_resources

        for dist in pkg_resources.working_set:  # pylint: disable=not-an-iterable
            module_path = dist.module_path or dist.location
            if not module_path:
                # Skip if no module path was found for pkg distribution
                continue

            if os.path.realpath(module_path) != os.getcwd():
                # add to nonlocal_package path only if it's not current directory
                self.nonlocal_package_path.add(module_path)

            self.pip_pkg_map[dist._key] = dist._version
            for mn in dist._get_metadata("top_level.txt"):
                if dist._key != "setuptools":
                    self.pip_module_map.setdefault(mn, []).append(
                        (dist._key, dist._version)
                    )
                else:
                    self.setuptools_module_set.add(mn)

        self.searched_modules = {}
        self.zip_modules: Dict[str, zipimport.zipimporter] = {}
        for m in pkgutil.iter_modules():
            if m.name not in self.searched_modules:
                if isinstance(m.module_finder, zipimport.zipimporter):
                    logger.info(f"Detected zipimporter {m.module_finder}")
                    path = m.module_finder.archive
                    self.zip_modules[path] = m.module_finder
                else:
                    path = m.module_finder.path
                is_local = self.is_local_path(path)
                self.searched_modules[m.name] = ModuleInfo(
                    m.name, path, is_local, m.ispkg
                )

    def verify_pkg(self, pkg_req):
        if pkg_req.name not in self.pip_pkg_map:
            # package does not exist in the current python session
            return EPP_PKG_NOT_EXIST

        if self.pip_pkg_map[pkg_req.name] not in pkg_req.specifier:
            # package version being used in the current python session does not meet
            # the specified package version requirement
            return EPP_PKG_VERSION_MISMATCH

        return EPP_NO_ERROR

    def seek_pip_packages(self, target_py_file_path):
        logger.debug('target py file path: %s', target_py_file_path)
        work = DepSeekWork(self, target_py_file_path)
        work.do()
        requirements = {}
        for _, pkg_info_list in work.dependencies.items():
            for pkg_name, pkg_version in pkg_info_list:
                requirements[pkg_name] = pkg_version

        return requirements, work.unknown_module_set

    def is_local_path(self, path):
        if path in self.nonlocal_package_path:
            return False

        dir_name = os.path.split(path)[1]

        if (
            "site-packages" in path
            or "anaconda" in path
            or path.endswith("packages")
            or dir_name == "bin"
            or dir_name.startswith("lib")
            or dir_name.startswith("python")
            or dir_name.startswith("plat")
        ):
            self.nonlocal_package_path.add(path)
            return False

        return True


class DepSeekWork(object):
    def __init__(self, module_manager, target_py_file_path):
        super(DepSeekWork, self).__init__()
        self.module_manager = module_manager
        self.target_py_file_path = target_py_file_path

        self.dependencies = {}
        self.unknown_module_set = set()
        self.parsed_module_set = set()

    def do(self):
        self.seek_in_file(self.target_py_file_path)

    def seek_in_file(self, file_path):
        try:
            with open(file_path) as f:
                content = f.read()
        except UnicodeDecodeError:
            with open(file_path, encoding='utf-8') as f:
                content = f.read()
        self.seek_in_source(content)

    def seek_in_source(self, content):
        # Extract all dependency modules by searching through the trees of the Python
        # abstract syntax grammar with Python's built-in ast module
        tree = ast.parse(content)
        import_set = set()
        for node in ast.walk(tree):
            if isinstance(node, ast.Import):
                for name in node.names:
                    import_set.add(name.name.partition(".")[0])
            elif isinstance(node, ast.ImportFrom):
                if node.module is not None and node.level == 0:
                    import_set.add(node.module.partition(".")[0])
        logger.debug("import set: %s", import_set)
        for module_name in import_set:
            # Avoid parsing BentoML when BentoML is imported from local source code repo
            if module_name == 'bentoml':
                continue
            if module_name in self.parsed_module_set:
                continue
            self.parsed_module_set.add(module_name)

            if module_name in self.module_manager.searched_modules:
                m = self.module_manager.searched_modules[module_name]
                if m.is_local:
                    # Recursively search dependencies in sub-modules
                    if m.path in self.module_manager.zip_modules:
                        self.seek_in_zip(m.path)
                    elif m.is_pkg:
                        self.seek_in_dir(os.path.join(m.path, m.name))
                    else:
                        self.seek_in_file(os.path.join(m.path, "{}.py".format(m.name)))
                else:
                    # check if the package has already been added to the list
                    if (
                        module_name in self.module_manager.pip_module_map
                        and module_name not in self.dependencies
                        and module_name not in self.module_manager.setuptools_module_set
                    ):
                        self.dependencies[
                            module_name
                        ] = self.module_manager.pip_module_map[module_name]
            else:
                if module_name in self.module_manager.pip_module_map:
                    if module_name not in self.dependencies:
                        # In some special cases, the pip-installed module can not
                        # be located in the searched_modules
                        self.dependencies[
                            module_name
                        ] = self.module_manager.pip_module_map[module_name]
                else:
                    if module_name not in sys.builtin_module_names:
                        self.unknown_module_set.add(module_name)

    def seek_in_dir(self, dir_path):
        for path, dir_list, file_list in os.walk(dir_path):
            for file_name in file_list:
                if not file_name.endswith(".py"):
                    continue
                self.seek_in_file(os.path.join(path, file_name))
            for dir_name in dir_list:
                if dir_name == '__pycache__':
                    continue
                self.seek_in_dir(os.path.join(path, dir_name))

    def seek_in_zip(self, zip_path):
        with zipfile.ZipFile(zip_path) as zf:
            for module_path in zf.infolist():
                filename = module_path.filename
                if filename.endswith('.py'):
                    logger.debug("Seeking modules in zip %s", filename)
                    content = self.module_manager.zip_modules[zip_path].get_source(
                        filename.replace('.py', '')
                    )
                    self.seek_in_source(content)



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from bentoml.adapters.base_input import BaseInputAdapter

ADAPTER_TYPE_TO_INPUT_TYPE = {
    "ClipperBytesInput": "bytes",
    "ClipperIntsInput": "ints",
    "ClipperFloatsInput": "floats",
    "ClipperDoublesInput": "doubles",
    "ClipperStringsInput": "strings",
}


class ClipperInput(BaseInputAdapter):
    """
    A special input adapter that should only be used when deploying BentoService
     with Clipper(http://clipper.ai/)

    ClipperInput are not regular InputAdapter, they don't work with REST
    API server nor BentoML CLI.
    """

    @property
    def pip_dependencies(self):
        # 'clipper_admin' package is only required on the machine deploying BentoService
        # to a clipper cluster, not required by the API Server itself
        return []

    def from_http_request(self, req):
        raise NotImplementedError(
            "ClipperInput does not support handling REST API prediction request"
        )

    def from_cli(self, cli_args):
        raise NotImplementedError(
            "ClipperInput is not supported to be used with BentoML CLI"
        )

    def from_inference_job(self, *args, **kwargs):
        raise NotImplementedError(
            "" "ClipperInput is not supported to be used with BentoML function calling"
        )

    def from_aws_lambda_event(self, event):
        raise NotImplementedError(
            "ClipperInput is not supported in AWS Lambda Deployment"
        )


# pylint: disable=abstract-method
class ClipperBytesInput(ClipperInput):
    """
    ClipperInput that deals with input type Bytes
    """


class ClipperFloatsInput(ClipperInput):
    """
    ClipperInput that deals with input type Floats
    """


class ClipperIntsInput(ClipperInput):
    """
    ClipperInput that deals with input type Ints
    """


class ClipperDoublesInput(ClipperInput):
    """
    ClipperInput that deals with input type Doubles
    """


class ClipperStringsInput(ClipperInput):
    """
    ClipperInput that deals with input type String
    """



# Copyright 2020 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import multiprocessing
import os

from deepmerge import always_merger
from schema import And, Optional, Or, Schema, SchemaError, Use
from simple_di import container, providers

from bentoml import __version__
from bentoml.configuration import expand_env_var, get_bentoml_deploy_version
from bentoml.exceptions import BentoMLConfigException
from bentoml.utils.ruamel_yaml import YAML


LOGGER = logging.getLogger(__name__)

YATAI_REPOSITORY_S3 = "s3"
YATAI_REPOSITORY_GCS = "gcs"
YATAI_REPOSITORY_FILE_SYSTEM = "file_system"
YATAI_REPOSITORY_TYPES = [
    YATAI_REPOSITORY_FILE_SYSTEM,
    YATAI_REPOSITORY_S3,
    YATAI_REPOSITORY_GCS,
]

SCHEMA = Schema(
    {
        "bento_bundle": {
            "deployment_version": Or(str, None),
            "default_docker_base_image": Or(str, None),
        },
        "bento_server": {
            "port": And(int, lambda port: port > 0),
            "workers": Or(And(int, lambda workers: workers > 0), None),
            "timeout": And(int, lambda timeout: timeout > 0),
            "max_request_size": And(int, lambda size: size > 0),
            "microbatch": {
                Optional("enabled", default=True): bool,
                "workers": Or(And(int, lambda workers: workers > 0), None),
                "max_batch_size": Or(And(int, lambda size: size > 0), None),
                "max_latency": Or(And(int, lambda latency: latency > 0), None),
            },
            "ngrok": {"enabled": bool},
            "swagger": {"enabled": bool},
            "metrics": {"enabled": bool, "namespace": str},
            "feedback": {"enabled": bool},
            "logging": {"level": str},
            "cors": {
                "enabled": bool,
                "access_control_allow_origin": Or(str, None),
                "access_control_allow_credentials": Or(bool, None),
                "access_control_allow_headers": Or([str], str, None),
                "access_control_allow_methods": Or([str], str, None),
                "access_control_max_age": Or(int, None),
                "access_control_expose_headers": Or([str], str, None),
            },
        },
        "logging": {
            "level": And(
                str,
                lambda level: level.isupper(),
                error="logging.level must be all upper case letters",
            ),
            "console": {"enabled": bool},
            "file": {"enabled": bool, "directory": Or(str, None)},
            "advanced": {"enabled": bool, "config": Or(dict, None)},
        },
        "tracing": {
            "type": Or(
                And(str, Use(str.lower), lambda s: s in ('zipkin', 'jaeger')), None
            ),
            Optional("zipkin"): {"url": Or(str, None)},
            Optional("jaeger"): {"address": Or(str, None), "port": Or(int, None)},
        },
        "adapters": {"image_input": {"default_extensions": [str]}},
        "yatai": {
            "remote": {
                "url": Or(str, None),
                "access_token": Or(str, None),
                "access_token_header": Or(str, None),
                "tls": {
                    "root_ca_cert": Or(str, None),
                    "client_key": Or(str, None),
                    "client_cert": Or(str, None),
                    "client_certificate_file": Or(str, None),
                },
            },
            "repository": {
                "type": And(
                    str,
                    lambda type: type in YATAI_REPOSITORY_TYPES,
                    error="yatai.repository.type must be one of %s"
                    % YATAI_REPOSITORY_TYPES,
                ),
                "file_system": {"directory": Or(str, None)},
                "s3": {
                    "url": Or(str, None),
                    "endpoint_url": Or(str, None),
                    "signature_version": Or(str, None),
                    "expiration": Or(int, None),
                },
                "gcs": {"url": Or(str, None), "expiration": Or(int, None)},
            },
            "database": {"url": Or(str, None)},
            "namespace": str,
            "logging": {"path": Or(str, None)},
        },
    }
)


class BentoMLConfiguration:
    def __init__(
        self,
        default_config_file: str = None,
        override_config_file: str = None,
        validate_schema: bool = True,
    ):
        # Default configuraiton
        if default_config_file is None:
            default_config_file = os.path.join(
                os.path.dirname(__file__), "default_configuration.yml"
            )

        with open(default_config_file, "rb") as f:
            self.config = YAML().load(f.read())

        if validate_schema:
            try:
                SCHEMA.validate(self.config)
            except SchemaError as e:
                raise BentoMLConfigException(
                    "Default configuration 'default_configuration.yml' does not"
                    " conform to the required schema."
                ) from e

        # User override configuration
        if override_config_file is not None:
            LOGGER.info("Applying user config override from %s" % override_config_file)
            if not os.path.exists(override_config_file):
                raise BentoMLConfigException(
                    f"Config file {override_config_file} not found"
                )

            with open(override_config_file, "rb") as f:
                override_config = YAML().load(f.read())
            always_merger.merge(self.config, override_config)

            if validate_schema:
                try:
                    SCHEMA.validate(self.config)
                except SchemaError as e:
                    raise BentoMLConfigException(
                        "Configuration after user override does not conform to"
                        " the required schema."
                    ) from e

    def override(self, keys: list, value):
        if keys is None:
            raise BentoMLConfigException("Configuration override key is None.")
        if len(keys) == 0:
            raise BentoMLConfigException("Configuration override key is empty.")
        if value is None:
            return

        c = self.config
        for key in keys[:-1]:
            if key not in c:
                raise BentoMLConfigException(
                    "Configuration override key is invalid, %s" % keys
                )
            c = c[key]
        c[keys[-1]] = value

        try:
            SCHEMA.validate(self.config)
        except SchemaError as e:
            raise BentoMLConfigException(
                "Configuration after applying override does not conform"
                " to the required schema, key=%s, value=%s." % (keys, value)
            ) from e

    def as_dict(self) -> dict:
        return self.config


def _get_cors_options(**kwargs):
    import aiohttp_cors

    filtered_kwargs = {k: v for k, v in kwargs.items() if v is not None}

    return aiohttp_cors.ResourceOptions(**filtered_kwargs)


@container
class BentoMLContainerClass:

    config = providers.Configuration()

    access_control_options = providers.Callable(
        _get_cors_options,
        allow_credentials=config.bento_server.cors.access_control_allow_credentials,
        expose_headers=config.bento_server.cors.access_control_expose_headers,
        allow_methods=config.bento_server.cors.access_control_allow_methods,
        allow_headers=config.bento_server.cors.access_control_allow_headers,
        max_age=config.bento_server.cors.access_control_max_age,
    )

    api_server_workers = providers.Callable(
        lambda workers: workers or (multiprocessing.cpu_count() // 2) + 1,
        config.bento_server.workers,
    )

    bentoml_home = providers.Callable(
        lambda: expand_env_var(
            os.environ.get("BENTOML_HOME", os.path.join("~", "bentoml"))
        )
    )

    prometheus_multiproc_dir = providers.Callable(
        os.path.join, bentoml_home, "prometheus_multiproc_dir",
    )

    bento_bundle_deployment_version = providers.Callable(
        get_bentoml_deploy_version,
        providers.Callable(
            lambda default, customized: customized or default,
            __version__.split('+')[0],
            config.bento_bundle.deployment_version,
        ),
    )

    yatai_database_url = providers.Callable(
        lambda default, customized: customized or default,
        providers.Callable(
            "sqlite:///{}".format,
            providers.Callable(os.path.join, bentoml_home, "storage.db"),
        ),
        config.yatai.database.url,
    )

    yatai_file_system_directory = providers.Callable(
        lambda default, customized: customized or default,
        providers.Callable(os.path.join, bentoml_home, "repository"),
        config.yatai.repository.file_system.directory,
    )

    yatai_tls_root_ca_cert = providers.Callable(
        lambda current, deprecated: current or deprecated,
        config.yatai.remote.tls.root_ca_cert,
        config.yatai.remote.tls.client_certificate_file,
    )

    logging_file_directory = providers.Callable(
        lambda default, customized: customized or default,
        providers.Callable(os.path.join, bentoml_home, "logs",),
        config.logging.file.directory,
    )

    yatai_logging_path = providers.Callable(
        lambda default, customized: customized or default,
        providers.Callable(
            os.path.join, logging_file_directory, "yatai_web_server.log"
        ),
        config.yatai.logging.path,
    )


BentoMLContainer = BentoMLContainerClass()



import logging
import multiprocessing
import os
import shutil
from timeit import default_timer

from flask import Request
from simple_di import Provide, inject

from bentoml.configuration.containers import BentoMLContainer


logger = logging.getLogger(__name__)


class InstrumentMiddleware:
    @inject
    def __init__(
        self,
        app,
        bento_service,
        namespace: str = Provide[
            BentoMLContainer.config.bento_server.metrics.namespace
        ],
    ):
        self.app = app
        self.bento_service = bento_service

        from prometheus_client import (
            Histogram,
            Counter,
            Gauge,
            CollectorRegistry,
        )

        service_name = self.bento_service.name
        self.collector_registry = CollectorRegistry()

        self.metrics_request_duration = Histogram(
            name=service_name + '_request_duration_seconds',
            documentation=service_name + " API HTTP request duration in seconds",
            namespace=namespace,
            labelnames=['endpoint', 'service_version', 'http_response_code'],
            registry=self.collector_registry,
        )
        self.metrics_request_total = Counter(
            name=service_name + "_request_total",
            documentation='Total number of HTTP requests',
            namespace=namespace,
            labelnames=['endpoint', 'service_version', 'http_response_code'],
            registry=self.collector_registry,
        )
        self.metrics_request_in_progress = Gauge(
            name=service_name + "_request_in_progress",
            documentation='Total number of HTTP requests in progress now',
            namespace=namespace,
            labelnames=['endpoint', 'service_version'],
            multiprocess_mode='livesum',
            registry=self.collector_registry,
        )

    def __call__(self, environ, start_response):
        req = Request(environ)
        endpoint = req.path
        start_time = default_timer()

        def start_response_wrapper(status, headers, exc_info=None):
            ret = start_response(status, headers, exc_info)
            status_code = int(status.split()[0])

            # instrument request total count
            self.metrics_request_total.labels(
                endpoint=endpoint,
                service_version=self.bento_service.version,
                http_response_code=status_code,
            ).inc()

            # instrument request duration
            total_time = max(default_timer() - start_time, 0)
            self.metrics_request_duration.labels(
                endpoint=endpoint,
                service_version=self.bento_service.version,
                http_response_code=status_code,
            ).observe(total_time)

            return ret

        with self.metrics_request_in_progress.labels(
            endpoint=endpoint, service_version=self.bento_service.version
        ).track_inprogress():
            return self.app(environ, start_response_wrapper)


@inject
def setup_prometheus_multiproc_dir(
    lock: multiprocessing.Lock = None,
    prometheus_multiproc_dir: str = Provide[BentoMLContainer.prometheus_multiproc_dir],
):
    """
    Set up prometheus_multiproc_dir for prometheus to work in multiprocess mode,
    which is required when working with Gunicorn server

    Warning: for this to work, prometheus_client library must be imported after
    this function is called. It relies on the os.environ['prometheus_multiproc_dir']
    to properly setup for multiprocess mode
    """
    if lock is not None:
        lock.acquire()

    try:
        logger.debug(
            "Setting up prometheus_multiproc_dir: %s", prometheus_multiproc_dir
        )
        # Wipe prometheus metrics directory between runs
        # https://github.com/prometheus/client_python#multiprocess-mode-gunicorn
        # Ignore errors so it does not fail when directory does not exist
        shutil.rmtree(prometheus_multiproc_dir, ignore_errors=True)
        os.makedirs(prometheus_multiproc_dir, exist_ok=True)

        os.environ['prometheus_multiproc_dir'] = prometheus_multiproc_dir
    finally:
        if lock is not None:
            lock.release()



"""create lock table

Revision ID: af18f0b83286
Revises: 719dd2aacc9a
Create Date: 2021-05-18 18:33:10.622364

"""
from alembic import op


# revision identifiers, used by Alembic.
from bentoml.yatai.db.stores.lock import Lock

revision = 'af18f0b83286'
down_revision = '719dd2aacc9a'
branch_labels = None
depends_on = None


def upgrade():
    bind = op.get_bind()
    Lock.__table__.create(bind)


def downgrade():
    op.drop_table('locks')



import os
import io
import tarfile
from bentoml.yatai.proto.repository_pb2 import UploadBentoRequest, DownloadBentoResponse


DEFAULT_FILE_CHUNK_SIZE = 1024 * 8


class FileStream:
    def __init__(self):
        self.buffer = io.BytesIO()
        self.offset = 0

    def write(self, s):
        self.buffer.write(s)
        self.offset += len(s)

    def tell(self):
        return self.offset

    def close(self):
        self.buffer.close()

    def read_value(self):
        try:
            return self.buffer.getvalue()
        finally:
            self.buffer.close()
            self.buffer = io.BytesIO()


class _BentoBundleStreamRequestsOrResponses:
    def __init__(
        self,
        bento_name,
        bento_version,
        bento_bundle_path,
        file_chunk_size=DEFAULT_FILE_CHUNK_SIZE,
        is_request=True,
    ):
        """
        Build a tar file chunk by chunk on the fly from a directory and then stream out
        either request or response for upload/download.

        Args:
            bento_name:
            bento_version:
            bento_bundle_path:
            file_chunk_size:
            is_request:
        """
        self.bento_name = bento_name
        self.bento_version = bento_version
        self.directory_path = bento_bundle_path
        self.file_chunk_size = file_chunk_size
        self.is_request = is_request
        self.sent_init_message = False
        self.out_stream = FileStream()
        self.tar = tarfile.TarFile(fileobj=self.out_stream, mode='w')

    @staticmethod
    def _stream_file_into_tar(tarinfo, tar, file_handler, buf_size):
        out = tar.fileobj

        # For iter(), the second argument, sentinel, is given, then object must
        # be a callable object. The iterator created in this case will
        # call object with no arguments for each call to its __next__()
        # method; if the value returned is equal to sentinel, StopIteration
        # will be raised, otherwise the value will be returned.
        for b in iter(lambda: file_handler.read(buf_size), b''):
            out.write(b)
            yield
        blocks, remainder = divmod(tarinfo.size, tarfile.BLOCKSIZE)
        if remainder > 0:
            out.write(tarfile.NUL * (tarfile.BLOCKSIZE - remainder))
            blocks += 1
        tar.offset += blocks * tarfile.BLOCKSIZE
        yield

    @staticmethod
    def _create_request_or_response(bento_name, bento_version, value, is_request):
        if is_request:
            return UploadBentoRequest(
                bento_name=bento_name, bento_version=bento_version, bento_bundle=value,
            )
        else:
            return DownloadBentoResponse(bento_bundle=value)

    def close(self):
        self.tar.close()
        self.out_stream.close()

    def __iter__(self):
        if self.is_request and not self.sent_init_message:
            self.sent_init_message = True
            yield UploadBentoRequest(
                bento_name=self.bento_name, bento_version=self.bento_version
            )
        # Include length of separator.
        prefix_len = len(self.directory_path) + len(os.path.sep)

        # Add the directory path to tar
        self.tar.add(name=self.directory_path, arcname='', recursive=False)

        # walk the directory and add to tarfile
        for path, dirs, files in os.walk(self.directory_path):
            arc_path = path[prefix_len:]

            # Add files to tar. Add the tar info and then stream the file into tar
            for f in files:
                file_path = os.path.join(path, f)
                with open(file_path, 'rb') as file:
                    # Reading the file info and generate tarinfo from that.
                    # No file content is added to the tar at this point
                    tar_info = self.tar.gettarinfo(
                        name=file_path, arcname=os.path.join(arc_path, f), fileobj=file,
                    )
                    self.tar.addfile(tar_info)
                    # stream the file content into the tar
                    for _ in self._stream_file_into_tar(
                        tar_info, self.tar, file, self.file_chunk_size
                    ):
                        # Stream out as tarfile
                        yield self._create_request_or_response(
                            self.bento_name,
                            self.bento_version,
                            self.out_stream.read_value(),
                            self.is_request,
                        )

            # Add the directory path to tar
            for dir_path in dirs:
                self.tar.add(
                    name=os.path.join(path, dir_path),
                    arcname=os.path.join(arc_path, dir_path),
                    recursive=False,
                )
            # Stream out directories info in the tar
            yield self._create_request_or_response(
                self.bento_name,
                self.bento_version,
                self.out_stream.read_value(),
                self.is_request,
            )
        self.tar.close()
        # Stream out any value that are left over.
        # In tarfile write mode, two finishing zero blocks are appended to the archive.
        # https://docs.python.org/3/library/tarfile.html?highlight=tarfile#tarfile.TarFile.close
        yield self._create_request_or_response(
            self.bento_name,
            self.bento_version,
            self.out_stream.read_value(),
            self.is_request,
        )
        self.out_stream.close()


class UploadBentoStreamRequests(_BentoBundleStreamRequestsOrResponses):
    def __init__(
        self,
        bento_name,
        bento_version,
        bento_bundle_path,
        file_chunk_size=DEFAULT_FILE_CHUNK_SIZE,
    ):
        super().__init__(
            bento_name=bento_name,
            bento_version=bento_version,
            bento_bundle_path=bento_bundle_path,
            file_chunk_size=file_chunk_size,
            is_request=True,
        )


class DownloadBentoStreamResponses(_BentoBundleStreamRequestsOrResponses):
    def __init__(
        self,
        bento_name,
        bento_version,
        bento_bundle_path,
        file_chunk_size=DEFAULT_FILE_CHUNK_SIZE,
    ):
        super().__init__(
            bento_name=bento_name,
            bento_version=bento_version,
            bento_bundle_path=bento_bundle_path,
            file_chunk_size=file_chunk_size,
            is_request=False,
        )



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0


# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import functools
import inspect
import itertools
import logging
import sys
from typing import Callable, Iterable, Iterator, Sequence

from bentoml.adapters import BaseInputAdapter, BaseOutputAdapter
from bentoml.exceptions import BentoMLConfigException
from bentoml.tracing import get_tracer
from bentoml.types import HTTPRequest, HTTPResponse, InferenceResult, InferenceTask
from bentoml.utils import cached_property


logger = logging.getLogger(__name__)
prediction_logger = logging.getLogger("bentoml.prediction")


class InferenceAPI(object):
    """
    InferenceAPI defines an inference call to the underlying model, including its input
    and output adapter, the user-defined API callback function, and configurations for
    working with the BentoML adaptive micro-batching mechanism
    """

    def __init__(
        self,
        service,
        name,
        doc,
        input_adapter: BaseInputAdapter,
        user_func: Callable,
        output_adapter: BaseOutputAdapter,
        mb_max_latency=10000,
        mb_max_batch_size=1000,
        batch=False,
        route=None,
    ):
        """
        :param service: ref to service containing this API
        :param name: API name
        :param doc: the user facing document of this inference API, default to the
            docstring of the inference API function
        :param input_adapter: A InputAdapter that transforms HTTP Request and/or
            CLI options into parameters for API func
        :param user_func: the user-defined API callback function, this is
            typically the 'predict' method on a model
        :param output_adapter: A OutputAdapter is an layer between result of user
            defined API callback function
            and final output in a variety of different forms,
            such as HTTP response, command line stdout or AWS Lambda event object.
        :param mb_max_latency: The latency goal of this inference API in milliseconds.
            Default: 10000.
        :param mb_max_batch_size: The maximum size of requests batch accepted by this
            inference API. This parameter governs the throughput/latency trade off, and
            avoids having large batches that exceed some resource constraint (e.g. GPU
            memory to hold the entire batch's data). Default: 1000.
        :param batch: If true, the user API functool would take a batch of input data
            a time.
        :param route: Specify HTTP URL route of this inference API. By default,
            `api.name` is used as the route.  This parameter can be used for customizing
            the URL route, e.g. `route="/api/v2/model_a/predict"`
        """
        self._service = service
        self._name = name
        self._input_adapter = input_adapter
        self._user_func = user_func
        self._output_adapter = output_adapter
        self.mb_max_latency = mb_max_latency
        self.mb_max_batch_size = mb_max_batch_size
        self.batch = batch
        self.route = name if route is None else route

        if not self.input_adapter.BATCH_MODE_SUPPORTED and batch:
            raise BentoMLConfigException(
                f"{input_adapter.__class__.__name__} does not support `batch=True`"
            )

        if not self.input_adapter.SINGLE_MODE_SUPPORTED and not batch:
            raise BentoMLConfigException(
                f"{input_adapter.__class__.__name__} does not support `batch=False`, "
                "its output passed to API functions could only be a batch of data."
            )

        if doc is None:
            # generate a default doc string for this inference API
            doc = (
                f"BentoService inference API '{self.name}', input: "
                f"'{type(input_adapter).__name__}', output: "
                f"'{type(output_adapter).__name__}'"
            )
        self._doc = doc

    @property
    def service(self):
        """
        :return: a reference to the BentoService serving this inference API
        """
        return self._service

    @property
    def name(self):
        """
        :return: the name of this inference API
        """
        return self._name

    @property
    def doc(self):
        """
        :return: user facing documentation of this inference API
        """
        return self._doc

    @property
    def input_adapter(self) -> BaseInputAdapter:
        """
        :return: the input adapter of this inference API
        """
        return self._input_adapter

    @property
    def output_adapter(self) -> BaseOutputAdapter:
        """
        :return: the output adapter of this inference API
        """
        return self._output_adapter

    @cached_property
    def user_func(self):
        """
        :return: user-defined inference API callback function
        """

        # allow user to define handlers without 'tasks' kwargs
        _sig = inspect.signature(self._user_func)
        if self.batch:
            append_arg = "tasks"
        else:
            append_arg = "task"
        try:
            _sig.bind_partial(**{append_arg: None})
            append_arg = None
        except TypeError:
            pass

        @functools.wraps(self._user_func)
        def wrapped_func(*args, **kwargs):
            with get_tracer().span(
                service_name=f"BentoService.{self.service.name}",
                span_name=f"InferenceAPI {self.name} user defined callback function",
            ):
                if append_arg and append_arg in kwargs:
                    tasks = kwargs.pop(append_arg)
                elif append_arg in kwargs:
                    tasks = kwargs[append_arg]
                else:
                    tasks = []
                try:
                    return self._user_func(*args, **kwargs)
                except Exception as e:  # pylint: disable=broad-except
                    logger.error("Error caught in API function:", exc_info=1)
                    if self.batch:
                        for task in tasks:
                            if not task.is_discarded:
                                task.discard(
                                    http_status=500,
                                    err_msg=f"Exception happened in API function: {e}",
                                )
                        return [None] * sum(
                            1 if t.batch is None else t.batch for t in tasks
                        )
                    else:
                        task = tasks
                        if not task.is_discarded:
                            task.discard(
                                http_status=500,
                                err_msg=f"Exception happened in API function: {e}",
                            )
                        return [None] * (1 if task.batch is None else task.batch)

        return wrapped_func

    @property
    def request_schema(self):
        """
        :return: the HTTP API request schema in OpenAPI/Swagger format
        """
        if self.input_adapter.custom_request_schema is None:
            schema = self.input_adapter.request_schema
        else:
            schema = self.input_adapter.custom_request_schema

        if schema.get('application/json'):
            schema.get('application/json')[
                'example'
            ] = self.input_adapter._http_input_example
        return schema

    def _filter_tasks(
        self, inf_tasks: Iterable[InferenceTask]
    ) -> Iterator[InferenceTask]:
        for task in inf_tasks:
            if task.is_discarded:
                continue
            try:
                self.input_adapter.validate_task(task)
                yield task
            except AssertionError as e:
                task.discard(http_status=400, err_msg=str(e))

    def infer(self, inf_tasks: Iterable[InferenceTask]) -> Sequence[InferenceResult]:
        inf_tasks = tuple(inf_tasks)

        # extract args
        user_args = self.input_adapter.extract_user_func_args(inf_tasks)
        filtered_tasks = tuple(t for t in inf_tasks if not t.is_discarded)

        # call user function
        if not self.batch:  # For single inputs
            user_return = []
            for task, legacy_user_args in zip(
                filtered_tasks,
                self.input_adapter.iter_batch_args(user_args, tasks=filtered_tasks),
            ):
                ret = self.user_func(*legacy_user_args, task=task)
                if task.is_discarded:
                    continue
                else:
                    user_return.append(ret)
            if (
                isinstance(user_return, (list, tuple))
                and len(user_return)
                and isinstance(user_return[0], InferenceResult)
            ):
                inf_results = user_return
            else:
                # pack return value
                filtered_tasks = tuple(t for t in inf_tasks if not t.is_discarded)
                inf_results = self.output_adapter.pack_user_func_return_value(
                    user_return, tasks=filtered_tasks
                )
        else:
            user_return = self.user_func(*user_args, tasks=filtered_tasks)
            if (
                isinstance(user_return, (list, tuple))
                and len(user_return)
                and isinstance(user_return[0], InferenceResult)
            ):
                inf_results = user_return
            else:
                # pack return value
                filtered_tasks = tuple(t for t in inf_tasks if not t.is_discarded)
                inf_results = self.output_adapter.pack_user_func_return_value(
                    user_return, tasks=filtered_tasks
                )

        full_results = InferenceResult.complete_discarded(inf_tasks, inf_results)

        log_data = dict(
            service_name=self.service.name if self.service else "",
            service_version=self.service.version if self.service else "",
            api=self.name,
        )
        for task, result in zip(inf_tasks, inf_results):
            prediction_logger.info(
                dict(
                    log_data,
                    task=task.to_json(),
                    result=result.to_json(),
                    request_id=task.task_id,
                )
            )

        return tuple(full_results)

    def handle_request(self, request: HTTPRequest) -> HTTPResponse:
        inf_task = self.input_adapter.from_http_request(request)
        results = self.infer((inf_task,))
        result = next(iter(results))
        response = self.output_adapter.to_http_response(result)
        response.headers['X-Request-Id'] = inf_task.task_id
        return response

    def handle_batch_request(self, requests: Sequence[HTTPRequest]):
        with get_tracer().span(
            service_name=f"BentoService.{self.service.name}",
            span_name=f"InferenceAPI {self.name} handle batch requests",
        ):
            inf_tasks = tuple(map(self.input_adapter.from_http_request, requests))
            results = self.infer(inf_tasks)
            responses = tuple(map(self.output_adapter.to_http_response, results))
            for inf_task, response in zip(inf_tasks, responses):
                response.headers['X-Request-Id'] = inf_task.task_id
            return responses

    def handle_cli(self, cli_args: Sequence[str]) -> int:
        parser = argparse.ArgumentParser()
        parser.add_argument("--max-batch-size", default=sys.maxsize, type=int)
        parsed_args, _ = parser.parse_known_args(cli_args)

        exit_code = 0

        tasks_iter = self.input_adapter.from_cli(tuple(cli_args))
        while True:
            tasks = tuple(itertools.islice(tasks_iter, parsed_args.max_batch_size))
            if not len(tasks):
                break
            results = self.infer(tasks)
            exit_code = exit_code or self.output_adapter.to_cli(results)

        return exit_code

    def handle_aws_lambda_event(self, event):
        inf_task = self.input_adapter.from_aws_lambda_event(event)
        result = next(iter(self.infer((inf_task,))))
        out_event = self.output_adapter.to_aws_lambda_event(result)
        if isinstance(out_event, dict) and "headers" in out_event:
            headers = out_event.get("headers", dict())
            headers["Access-Control-Allow-Origin"] = "*"  # TODO: make it configurable
            out_event["headers"] = headers
        return out_event



import importlib
import os

from bentoml.exceptions import MissingDependencyException
from bentoml.service.env import BentoServiceEnv
from bentoml.service.artifacts import BentoServiceArtifact

EVALML_MODEL_PICKLE_EXTENTION = ".pkl"


class EvalMLModelArtifact(BentoServiceArtifact):
    """
    Artifact class  for saving/loading EvalML models

    Args:
        name (str): Name for the artifact

    Raises:
        MissingDependencyException: evalml package is required for EvalMLModelArtifact

    Example usage:

    >>> # import the EvalML base pipeline class corresponding to your ML task
    >>> from evalml.pipelines import BinaryClassificationPipeline
    >>> # create an EvalML pipeline instance with desired components
    >>> model_to_save = BinaryClassificationPipeline(
    >>>     ['Imputer', 'One Hot Encoder', 'Random Forest Classifier'])
    >>> # Train model with data using model_to_save.fit(X, y)
    >>>
    >>> import bentoml
    >>> from bentoml.frameworks.sklearn import EvalMLModelArtifact
    >>> from bentoml.adapters import DataframeInput
    >>>
    >>> @bentoml.env(infer_pip_packages=True)
    >>> @bentoml.artifacts([EvalMLModelArtifact('EvalML pipeline')])
    >>> class EvalMLModelService(bentoml.BentoService):
    >>>
    >>>     @bentoml.api(input=DataframeInput(), batch=True)
    >>>     def predict(self, df):
    >>>         result = self.artifacts.model.predict(df)
    >>>         return result
    >>>
    >>> svc = EvalMLModelService()
    >>>
    >>> # Pack directly with EvalML model object
    >>> svc.pack('model', model_to_save)
    >>> svc.save()
    """

    def __init__(self, name):
        super().__init__(name)

        self._model = None

    def _validate_package(self):
        try:
            importlib.import_module('evalml')
        except ImportError:
            raise MissingDependencyException(
                "Package 'evalml' is required to use EvalMLModelArtifact"
            )

    def _model_file_path(self, base_path):
        return os.path.join(base_path, self.name + EVALML_MODEL_PICKLE_EXTENTION)

    def pack(self, evalml_model, metadata=None):  # pylint:disable=arguments-renamed
        self._validate_package()
        self._model = evalml_model
        return self

    def load(self, path):
        self._validate_package()
        model_file_path = self._model_file_path(path)
        evalml_pipelines_module = importlib.import_module('evalml.pipelines')
        model = evalml_pipelines_module.PipelineBase.load(model_file_path)
        self.pack(model)
        return model

    def get(self):
        return self._model

    def save(self, dst):
        self._validate_package()
        model_file_path = self._model_file_path(dst)
        self._model.save(model_file_path)

    def set_dependencies(self, env: BentoServiceEnv):
        if env._infer_pip_packages:
            env.add_pip_packages(['evalml'])



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click
import logging

from bentoml.utils.lazy_loader import LazyLoader
from bentoml.cli.click_utils import (
    BentoMLCommandGroup,
    _echo,
    CLI_COLOR_SUCCESS,
    parse_yaml_file_callback,
)
from bentoml.yatai.deployment import ALL_NAMESPACE_TAG
from bentoml.utils import status_pb_to_error_code_and_message
from bentoml.exceptions import CLIException
from bentoml.cli.utils import Spinner, _print_deployment_info, _print_deployments_info
from bentoml.utils import get_default_yatai_client

yatai_proto = LazyLoader('yatai_proto', globals(), 'bentoml.yatai.proto')

# pylint: disable=unused-variable

logger = logging.getLogger(__name__)


DEFAULT_SAGEMAKER_INSTANCE_TYPE = 'ml.m4.xlarge'
DEFAULT_SAGEMAKER_INSTANCE_COUNT = 1


def get_deployment_sub_command():
    # pylint: disable=unused-variable

    @click.group(
        help='Commands for managing and operating BentoService deployments',
        cls=BentoMLCommandGroup,
    )
    def deployment():
        pass

    @deployment.command(help='Create BentoService deployment from yaml file')
    @click.option(
        '-f',
        '--file',
        'deployment_yaml',
        type=click.File('r'),
        required=True,
        callback=parse_yaml_file_callback,
    )
    @click.option('-o', '--output', type=click.Choice(['json', 'yaml']), default='json')
    @click.option(
        '--wait/--no-wait',
        default=True,
        help='Wait for apply action to complete or encounter an error.'
        'If set to no-wait, CLI will return immediately. The default value is wait',
    )
    def create(deployment_yaml, output, wait):
        yatai_client = get_default_yatai_client()
        deployment_name = deployment_yaml.get('name')
        with Spinner('Creating deployment '):
            result = yatai_client.deployment.create(deployment_yaml, wait)
        if result.status.status_code != yatai_proto.status_pb2.Status.OK:
            error_code, error_message = status_pb_to_error_code_and_message(
                result.status
            )
            raise CLIException(f'{error_code}:{error_message}')
        _echo(
            f'Successfully created deployment {deployment_name}', CLI_COLOR_SUCCESS,
        )
        _print_deployment_info(result.deployment, output)

    @deployment.command(help='Apply BentoService deployment from yaml file')
    @click.option(
        '-f',
        '--file',
        'deployment_yaml',
        type=click.File('r'),
        required=True,
        callback=parse_yaml_file_callback,
    )
    @click.option('-o', '--output', type=click.Choice(['json', 'yaml']), default='json')
    @click.option(
        '--wait/--no-wait',
        default=True,
        help='Wait for apply action to complete or encounter an error.'
        'If set to no-wait, CLI will return immediately. The default value is wait',
    )
    def apply(deployment_yaml, output, wait):
        deployment_name = deployment_yaml.get('name')
        yatai_client = get_default_yatai_client()
        with Spinner('Applying deployment'):
            result = yatai_client.deployment.apply(deployment_yaml, wait)
        if result.status.status_code != yatai_proto.status_pb2.Status.OK:
            error_code, error_message = status_pb_to_error_code_and_message(
                result.status
            )
            raise CLIException(f'{error_code}:{error_message}')
        _echo(
            f'Successfully applied deployment {deployment_name}', CLI_COLOR_SUCCESS,
        )
        _print_deployment_info(result.deployment, output)

    @deployment.command(help='Delete deployment')
    @click.argument('name', type=click.STRING)
    @click.option(
        '-n',
        '--namespace',
        type=click.STRING,
        help='Deployment namespace managed by BentoML, default value is "default" '
        'which can be changed in BentoML configuration file',
    )
    @click.option(
        '--force',
        is_flag=True,
        help='force delete the deployment record in database and '
        'ignore errors when deleting cloud resources',
    )
    def delete(name, namespace, force):
        yatai_client = get_default_yatai_client()
        get_deployment_result = yatai_client.deployment.get(namespace, name)
        if get_deployment_result.status.status_code != yatai_proto.status_pb2.Status.OK:
            error_code, error_message = status_pb_to_error_code_and_message(
                get_deployment_result.status
            )
            raise CLIException(f'{error_code}:{error_message}')
        result = yatai_client.deployment.delete(name, namespace, force)
        if result.status.status_code != yatai_proto.status_pb2.Status.OK:
            error_code, error_message = status_pb_to_error_code_and_message(
                result.status
            )
            raise CLIException(f'{error_code}:{error_message}')
        _echo('Successfully deleted deployment "{}"'.format(name), CLI_COLOR_SUCCESS)

    @deployment.command(help='Get deployment information')
    @click.argument('name', type=click.STRING)
    @click.option(
        '-n',
        '--namespace',
        type=click.STRING,
        help='Deployment namespace managed by BentoML, default value is "dev" which '
        'can be changed in BentoML configuration file',
    )
    @click.option('-o', '--output', type=click.Choice(['json', 'yaml']), default='json')
    def get(name, namespace, output):
        yatai_client = get_default_yatai_client()
        get_result = yatai_client.deployment.get(namespace, name)
        if get_result.status.status_code != yatai_proto.status_pb2.Status.OK:
            error_code, error_message = status_pb_to_error_code_and_message(
                get_result.status
            )
            raise CLIException(f'{error_code}:{error_message}')
        describe_result = yatai_client.deployment.describe(
            namespace=namespace, name=name
        )
        if describe_result.status.status_code != yatai_proto.status_pb2.Status.OK:
            error_code, error_message = status_pb_to_error_code_and_message(
                describe_result.status
            )
            raise CLIException(f'{error_code}:{error_message}')
        get_result.deployment.state.CopyFrom(describe_result.state)
        _print_deployment_info(get_result.deployment, output)

    @deployment.command(name='list', help='List deployments')
    @click.option(
        '-n',
        '--namespace',
        type=click.STRING,
        help='Deployment namespace managed by BentoML, default value is "dev" which '
        'can be changed in BentoML configuration file',
        default=ALL_NAMESPACE_TAG,
    )
    @click.option(
        '-p', '--platform', type=click.Choice(['sagemaker', 'lambda']), help='platform',
    )
    @click.option(
        '--limit',
        type=click.INT,
        help='The maximum amount of deployments to be listed at once',
    )
    @click.option(
        '--labels',
        type=click.STRING,
        help="Label query to filter deployments, supports '=', '!=', 'IN', 'NotIn', "
        "'Exists', and 'DoesNotExist'. (e.g. key1=value1, key2!=value2, key3 "
        "In (value3, value3a), key4 DoesNotExist)",
    )
    @click.option(
        '--order-by', type=click.Choice(['created_at', 'name']), default='created_at',
    )
    @click.option(
        '--asc/--desc',
        default=False,
        help='Ascending or descending order for list deployments',
    )
    @click.option(
        '-o',
        '--output',
        type=click.Choice(['json', 'yaml', 'table', 'wide']),
        default='table',
    )
    def list_deployments(namespace, platform, limit, labels, order_by, asc, output):
        yatai_client = get_default_yatai_client()
        list_result = yatai_client.deployment.list(
            limit=limit,
            labels=labels,
            namespace=namespace,
            operator=platform,
            order_by=order_by,
            ascending_order=asc,
        )
        if list_result.status.status_code != yatai_proto.status_pb2.Status.OK:
            error_code, error_message = status_pb_to_error_code_and_message(
                list_result.status
            )
            raise CLIException(f'{error_code}:{error_message}')
        _print_deployments_info(list_result.deployments, output)

    return deployment



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import multiprocessing
from typing import Optional

from simple_di import Provide, inject

from bentoml.configuration.containers import BentoMLContainer
from bentoml.marshal.marshal import MarshalService
from bentoml.server.instruments import setup_prometheus_multiproc_dir

marshal_logger = logging.getLogger("bentoml.marshal")


@inject
def gunicorn_marshal_server(
    default_workers: int = Provide[
        BentoMLContainer.config.bento_server.microbatch.workers
    ],
    default_timeout: int = Provide[BentoMLContainer.config.bento_server.timeout],
    default_outbound_workers: int = Provide[BentoMLContainer.api_server_workers],
    default_max_request_size: int = Provide[
        BentoMLContainer.config.bento_server.max_request_size
    ],
    default_port: int = Provide[BentoMLContainer.config.bento_server.port],
    default_mb_max_batch_size: int = Provide[
        BentoMLContainer.config.bento_server.microbatch.max_batch_size
    ],
    default_mb_max_latency: int = Provide[
        BentoMLContainer.config.bento_server.microbatch.max_latency
    ],
    default_loglevel: str = Provide[BentoMLContainer.config.bento_server.logging.level],
):
    from gunicorn.app.base import Application

    class GunicornMarshalServer(Application):  # pylint: disable=abstract-method
        @inject
        def __init__(
            self,
            bundle_path,
            outbound_host,
            outbound_port,
            workers: int = default_workers,
            timeout: int = default_timeout,
            outbound_workers: int = default_outbound_workers,
            max_request_size: int = default_max_request_size,
            port: int = default_port,
            mb_max_batch_size: int = default_mb_max_batch_size,
            mb_max_latency: int = default_mb_max_latency,
            prometheus_lock: Optional[multiprocessing.Lock] = None,
            loglevel: str = default_loglevel,
        ):
            self.bento_service_bundle_path = bundle_path

            self.port = port
            self.options = {
                "bind": "%s:%s" % ("0.0.0.0", self.port),
                "timeout": timeout,
                "limit_request_line": max_request_size,
                "loglevel": loglevel.upper(),
                "worker_class": "aiohttp.worker.GunicornWebWorker",
            }
            if workers:
                self.options['workers'] = workers
            self.prometheus_lock = prometheus_lock

            self.outbound_port = outbound_port
            self.outbound_host = outbound_host
            self.outbound_workers = outbound_workers
            self.mb_max_batch_size = mb_max_batch_size
            self.mb_max_latency = mb_max_latency

            super().__init__()

        def load_config(self):
            self.load_config_from_file("python:bentoml.server.gunicorn_config")

            # override config with self.options
            gunicorn_config = dict(
                [
                    (key, value)
                    for key, value in self.options.items()
                    if key in self.cfg.settings and value is not None
                ]
            )
            for key, value in gunicorn_config.items():
                self.cfg.set(key.lower(), value)

        def load(self):
            server = MarshalService(
                self.bento_service_bundle_path,
                self.outbound_host,
                self.outbound_port,
                outbound_workers=self.outbound_workers,
                mb_max_batch_size=self.mb_max_batch_size,
                mb_max_latency=self.mb_max_latency,
            )
            return server.make_app()

        def run(self):
            setup_prometheus_multiproc_dir(self.prometheus_lock)
            marshal_logger.info("Running micro batch service on :%d", self.port)
            super().run()

    return GunicornMarshalServer



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import azure.functions as func  # pylint: disable=import-error, no-name-in-module

from bentoml.server.api_server import BentoAPIServer
from bentoml.saved_bundle import load_from_dir

bento_path = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
svc = load_from_dir(bento_path)

bento_server = BentoAPIServer(svc)


def main(req: func.HttpRequest, context: func.Context) -> func.HttpResponse:
    return func.WsgiMiddleware(bento_server.app.wsgi_app).handle(req, context)



import re
import logging
from bentoml.exceptions import YataiDeploymentException

logger = logging.getLogger(__name__)


def to_valid_docker_image_name(name):
    # https://docs.docker.com/engine/reference/commandline/tag/#extended-description
    return name.lower().strip("._-")


def to_valid_docker_image_version(version):
    # https://docs.docker.com/engine/reference/commandline/tag/#extended-description
    return version.encode("ascii", errors="ignore").decode().lstrip(".-")[:128]


def validate_tag(ctx, param, tag):  # pylint: disable=unused-argument
    if tag is None:
        return tag

    if ":" in tag:
        name, version = tag.split(":")[:2]
    else:
        name, version = tag, None

    valid_name_pattern = re.compile(
        r"""
        ^(
        [a-z0-9]+      # alphanumeric
        (.|_{1,2}|-+)? # seperators
        )*$
        """,
        re.VERBOSE,
    )
    valid_version_pattern = re.compile(
        r"""
        ^
        [a-zA-Z0-9] # cant start with .-
        [ -~]{,127} # ascii match rest, cap at 128
        $
        """,
        re.VERBOSE,
    )

    if not valid_name_pattern.match(name):
        raise YataiDeploymentException(
            f"Provided Docker Image tag {tag} is invalid. "
            "Name components may contain lowercase letters, digits "
            "and separators. A separator is defined as a period, "
            "one or two underscores, or one or more dashes."
        )
    if version and not valid_version_pattern.match(version):
        raise YataiDeploymentException(
            f"Provided Docker Image tag {tag} is invalid. "
            "A tag name must be valid ASCII and may contain "
            "lowercase and uppercase letters, digits, underscores, "
            "periods and dashes. A tag name may not start with a period "
            "or a dash and may contain a maximum of 128 characters."
        )
    return tag



import os

from bentoml.exceptions import MissingDependencyException
from bentoml.service.artifacts import BentoServiceArtifact
from bentoml.service.env import BentoServiceEnv


class FasttextModelArtifact(BentoServiceArtifact):
    """
    Artifact class for saving and loading fasttext models

    Args:
        name (str): Name for the artifact

    Raises:
        MissingDependencyError: fasttext package is required for FasttextModelArtifact

    Example usage:

    >>> import fasttext
    >>> # prepare training data and store to file
    >>> training_data_file = 'training-data-file.train'
    >>> model = fasttext.train_supervised(input=training_data_file)
    >>>
    >>> import bentoml
    >>> from bentoml.adapters JsonInput
    >>> from bentoml.frameworks.fasttext import FasttextModelArtifact
    >>>
    >>> @bentoml.env(infer_pip_packages=True)
    >>> @bentoml.artifacts([FasttextModelArtifact('model')])
    >>> class FasttextModelService(bentoml.BentoService):
    >>>
    >>>     @bentoml.api(input=JsonInput(), batch=False)
    >>>     def predict(self, parsed_json):
    >>>         # K is the number of labels that successfully were predicted,
    >>>         # among all the real labels
    >>>         return self.artifacts.model.predict(parsed_json['text'], k=5)
    >>>
    >>> svc = FasttextModelService()
    >>> svc.pack('model', model)
    """

    def __init__(self, name: str):
        super().__init__(name)

        self._model = None

    def set_dependencies(self, env: BentoServiceEnv):
        if env._infer_pip_packages:
            env.add_pip_packages(["fasttext"])

    def _model_file_path(self, base_path):
        return os.path.join(base_path, self.name)

    def pack(self, fasttext_model, metadata=None):  # pylint:disable=arguments-renamed
        try:
            import fasttext  # noqa # pylint: disable=unused-import
        except ImportError:
            raise MissingDependencyException(
                "fasttext package is required to use FasttextModelArtifact"
            )
        self._model = fasttext_model
        return self

    def load(self, path):
        try:
            import fasttext  # noqa # pylint: disable=unused-import
        except ImportError:
            raise MissingDependencyException(
                "fasttext package is required to use FasttextModelArtifact"
            )

        model = fasttext.load_model(self._model_file_path(path))
        return self.pack(model)

    def get(self):
        return self._model

    def save(self, dst):
        self._model.save_model(self._model_file_path(dst))



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import time
import os

import click
import functools
import logging

from click import ClickException

from bentoml import configure_logging
from bentoml.configuration import set_debug_mode
from bentoml.exceptions import BentoMLException
from bentoml.utils.ruamel_yaml import YAML
from bentoml.utils.usage_stats import track

# Available CLI colors for _echo:
#
# _ansi_colors = {
#     'black': 30,
#     'red': 31,
#     'green': 32,
#     'yellow': 33,
#     'blue': 34,
#     'magenta': 35,
#     'cyan': 36,
#     'white': 37,
#     'reset': 39,
#     'bright_black': 90,
#     'bright_red': 91,
#     'bright_green': 92,
#     'bright_yellow': 93,
#     'bright_blue': 94,
#     'bright_magenta': 95,
#     'bright_cyan': 96,
#     'bright_white': 97,
# }

CLI_COLOR_SUCCESS = "green"
CLI_COLOR_ERROR = "red"
CLI_COLOR_WARNING = "yellow"

logger = logging.getLogger(__name__)
TRACK_CLI_EVENT_NAME = 'bentoml-cli'


def _echo(message, color="reset"):
    click.secho(message, fg=color)


class BentoMLCommandGroup(click.Group):
    """Click command class customized for BentoML cli, allow specifying a default
    command for each group defined
    """

    NUMBER_OF_COMMON_PARAMS = 2

    @staticmethod
    def bentoml_common_params(func):
        @click.option(
            '-q',
            '--quiet',
            is_flag=True,
            default=False,
            help="Hide all warnings and info logs",
        )
        @click.option(
            '--verbose',
            '--debug',
            is_flag=True,
            default=False,
            help="Show debug logs when running the command",
        )
        @click.option(
            '--do-not-track',
            is_flag=True,
            default=False,
            envvar="BENTOML_DO_NOT_TRACK",
            help="Specify the option to not track usage.",
        )
        @functools.wraps(func)
        def wrapper(quiet, verbose, *args, **kwargs):
            if quiet:
                configure_logging(logging_level=logging.ERROR)
                if verbose:
                    logger.warning(
                        "The bentoml command option `--verbose/--debug` is ignored when"
                        "the `--quiet` flag is also in use"
                    )
            elif verbose:
                set_debug_mode(True)

            return func(*args, **kwargs)

        return wrapper

    @staticmethod
    def bentoml_track_usage(func, cmd_group, **kwargs):
        command_name = kwargs.get('name', func.__name__)

        @functools.wraps(func)
        def wrapper(do_not_track: bool, *args, **kwargs):
            if do_not_track:
                os.environ["BENTOML_DO_NOT_TRACK"] = str(True)
                logger.debug(
                    "Executing '%s' command without usage tracking.", command_name
                )

            track_properties = {
                'command_group': cmd_group.name,
                'command': command_name,
            }
            start_time = time.time()
            try:
                return_value = func(*args, **kwargs)
                track_properties['duration'] = time.time() - start_time
                track_properties['return_code'] = 0
                track(TRACK_CLI_EVENT_NAME, track_properties)
                return return_value
            except BaseException as e:
                track_properties['duration'] = time.time() - start_time
                track_properties['error_type'] = type(e).__name__
                track_properties['error_message'] = str(e)
                track_properties['return_code'] = 1
                if type(e) == KeyboardInterrupt:
                    track_properties['return_code'] = 2
                track(TRACK_CLI_EVENT_NAME, track_properties)
                raise

        return wrapper

    @staticmethod
    def raise_click_exception(func, cmd_group, **kwargs):
        command_name = kwargs.get('name', func.__name__)

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except BentoMLException as e:
                msg = f'{cmd_group.name} {command_name} failed: {str(e)}'
                raise ClickException(click.style(msg, fg='red'))

        return wrapper

    def command(self, *args, **kwargs):
        def wrapper(func):
            # add common parameters to command
            func = BentoMLCommandGroup.bentoml_common_params(func)
            # Send tracking events before command finish.
            func = BentoMLCommandGroup.bentoml_track_usage(func, self, **kwargs)
            # If BentoMLException raise ClickException instead before exit
            func = BentoMLCommandGroup.raise_click_exception(func, self, **kwargs)

            # move common parameters to end of the parameters list
            func.__click_params__ = (
                func.__click_params__[-self.NUMBER_OF_COMMON_PARAMS :]
                + func.__click_params__[: -self.NUMBER_OF_COMMON_PARAMS]
            )
            return super(BentoMLCommandGroup, self).command(*args, **kwargs)(func)

        return wrapper


def conditional_argument(condition, *param_decls, **attrs):
    """
    Attaches an argument to the command only when condition is True
    """

    def decorator(f):
        if condition:
            f = click.argument(*param_decls, **attrs)(f)
        return f

    return decorator


def _is_valid_bento_tag(value):
    return re.match(r"^[A-Za-z_][A-Za-z_0-9]*:[A-Za-z0-9.+-_]*$", value) is not None


def _is_valid_bento_name(value):
    return re.match(r"^[A-Za-z_0-9]*$", value) is not None


def parse_bento_tag_callback(ctx, param, value):  # pylint: disable=unused-argument
    if param.required and not _is_valid_bento_tag(value):
        raise click.BadParameter(
            "Bad formatting. Please present in BentoName:Version, for example "
            "iris_classifier:v1.2.0"
        )
    return value


def parse_labels_callback(ctx, param, value):  # pylint: disable=unused-argument
    if not value:
        return value

    parsed_labels = {}
    label_list = value.split(',')
    for label in label_list:
        if ':' not in label:
            raise click.BadParameter(
                f'Bad formatting for label {label}. '
                f'Please present label in key:value format'
            )
        label_key, label_value = label.split(':')
        parsed_labels[label_key] = label_value

    return parsed_labels


def validate_labels_query_callback(
    ctx, param, value
):  # pylint: disable=unused-argument
    if not value:
        return value

    labels = value.split(',')
    for label in labels:
        if '=' not in label:
            raise click.BadParameter(
                f'Bad formatting for label {label}. '
                f'Please present labels query in key=value format'
            )
    return value


def parse_yaml_file_callback(ctx, param, value):  # pylint: disable=unused-argument
    yaml = YAML()
    yml_content = value.read()
    try:
        return yaml.load(yml_content)
    except Exception:
        raise click.BadParameter(
            'Input value is not recognizable yaml file or yaml content'
        )



import importlib
import types
import logging

logger = logging.getLogger(__name__)


class LazyLoader(types.ModuleType):
    """LazyLoader module borrowed from Tensorflow
  https://github.com/tensorflow/tensorflow/blob/v2.2.0/tensorflow/python/util/lazy_loader.py

  Lazily import a module, mainly to avoid pulling in large dependencies.
  `contrib`, and `ffmpeg` are examples of modules that are large and not always
  needed, and this allows them to only be loaded when they are used.
  """

    # The lint error here is incorrect.
    def __init__(self, local_name, parent_module_globals, name, warning=None):
        self._local_name = local_name
        self._parent_module_globals = parent_module_globals
        self._warning = warning

        super(LazyLoader, self).__init__(name)

    def _load(self):
        """Load the module and insert it into the parent's globals."""
        # Import the target module and insert it into the parent's namespace
        module = importlib.import_module(self.__name__)
        self._parent_module_globals[self._local_name] = module

        # Emit a warning if one was specified
        if self._warning:
            logger.warning(self._warning)
            # Make sure to only warn once.
            self._warning = None

        # Update this object's dict so that if someone keeps a reference to the
        #   LazyLoader, lookups are efficient (__getattr__ is only called on lookups
        #   that fail).
        self.__dict__.update(module.__dict__)

        return module

    def __getattr__(self, item):
        module = self._load()
        return getattr(module, item)

    def __dir__(self):
        module = self._load()
        return dir(module)



import os

from bentoml.service.env import BentoServiceEnv

from bentoml.exceptions import MissingDependencyException
from bentoml.service.artifacts import BentoServiceArtifact


DEFAULT_PICKLE_EXTENSION = ".pkl"


def _import_joblib_module():
    try:
        import joblib
    except ImportError:
        joblib = None

    if joblib is None:
        try:
            from sklearn.externals import joblib
        except ImportError:
            pass

    if joblib is None:
        raise MissingDependencyException(
            "sklearn module is required to use SklearnModelArtifact"
        )

    return joblib


class SklearnModelArtifact(BentoServiceArtifact):
    """
    Artifact class for saving/loading scikit learn models using sklearn.externals.joblib

    Args:
        name (str): Name for the artifact
        pickle_extension (str): The extension format for pickled file

    Raises:
        MissingDependencyException: sklearn package is required for SklearnModelArtifact

    Example usage:

    >>> from sklearn import svm
    >>>
    >>> model_to_save = svm.SVC(gamma='scale')
    >>> # ... training model, etc.
    >>>
    >>> import bentoml
    >>> from bentoml.frameworks.sklearn import SklearnModelArtifact
    >>> from bentoml.adapters import DataframeInput
    >>>
    >>> @bentoml.env(infer_pip_packages=True)
    >>> @bentoml.artifacts([SklearnModelArtifact('model')])
    >>> class SklearnModelService(bentoml.BentoService):
    >>>
    >>>     @bentoml.api(input=DataframeInput(), batch=True)
    >>>     def predict(self, df):
    >>>         result = self.artifacts.model.predict(df)
    >>>         return result
    >>>
    >>> svc = SklearnModelService()
    >>>
    >>> # Pack directly with sklearn model object
    >>> svc.pack('model', model_to_save)
    >>> svc.save()
    """

    def __init__(self, name):
        super().__init__(name)
        self._model = None

    def _model_file_path(self, base_path):
        return os.path.join(base_path, self.name + DEFAULT_PICKLE_EXTENSION)

    def pack(self, sklearn_model, metadata=None):  # pylint:disable=arguments-renamed
        self._model = sklearn_model
        return self

    def load(self, path):
        joblib = _import_joblib_module()

        model_file_path = self._model_file_path(path)
        sklearn_model = joblib.load(model_file_path, mmap_mode='r')
        return self.pack(sklearn_model)

    def get(self):
        return self._model

    def save(self, dst):
        joblib = _import_joblib_module()
        joblib.dump(self._model, self._model_file_path(dst))

    def set_dependencies(self, env: BentoServiceEnv):
        if env._infer_pip_packages:
            env.add_pip_packages(['scikit-learn'])



# Copyright 2017 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Interceptor that adds headers to outgoing requests."""

import collections

import grpc

from bentoml.yatai.client.interceptor import generic_client_interceptor


class _ClientCallDetails(
    collections.namedtuple(
        '_ClientCallDetails', ('method', 'timeout', 'metadata', 'credentials')
    ),
    grpc.ClientCallDetails,
):
    pass


def header_adder_interceptor(header, value):
    def intercept_call(
        client_call_details, request_iterator, request_streaming, response_streaming
    ):
        # pylint: disable=unused-argument
        metadata = []
        if client_call_details.metadata is not None:
            metadata = list(client_call_details.metadata)
        metadata.append((header, value,))
        client_call_details = _ClientCallDetails(
            client_call_details.method,
            client_call_details.timeout,
            metadata,
            client_call_details.credentials,
        )
        return client_call_details, request_iterator, None

    return generic_client_interceptor.create(intercept_call)



"""update lock table column

Revision ID: 370a2bf2e594
Revises: af18f0b83286
Create Date: 2021-06-15 13:41:18.015802

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '370a2bf2e594'
down_revision = 'af18f0b83286'
branch_labels = None
depends_on = None


def upgrade():
    # SQLite does not support making modifications to existing columns.
    # Alembic "batch mode" is designed to overcome this limitation.
    # https://alembic.sqlalchemy.org/en/latest/batch.html#batch-mode-with-autogenerate
    with op.batch_alter_table('locks') as batch_op:
        batch_op.alter_column(
            column_name='resource_id', type_=sa.String, existing_type=sa.INTEGER
        )


def downgrade():
    with op.batch_alter_table('locks') as batch_op:
        batch_op.alter_column(
            column_name='resource_id', type_=sa.INTEGER, existing_type=sa.String
        )



from .pickle import PickleArtifact  # noqa # pylint: disable=unused-import
from .json_file import JSONArtifact  # noqa # pylint: disable=unused-import
from .text_file import TextFileArtifact  # noqa # pylint: disable=unused-import



import os
import shutil

from bentoml.exceptions import MissingDependencyException
from bentoml.service.artifacts import BentoServiceArtifact
from bentoml.service.env import BentoServiceEnv


class H2oModelArtifact(BentoServiceArtifact):
    """
    Artifact class for saving and loading h2o models using h2o.save_model and
    h2o.load_model

    Args:
        name (str): Name for this h2o artifact..

    Raises:
        MissingDependencyException: h2o package is required to use H2o model artifact

    Example usage:

    >>> import h2o
    >>> h2o.init()
    >>>
    >>> from h2o.estimators.deeplearning import H2ODeepLearningEstimator
    >>> model_to_save = H2ODeepLearningEstimator(...)
    >>> # train model with data
    >>> data = h2o.import_file(...)
    >>> model_to_save.train(...)
    >>>
    >>> import bentoml
    >>> from bentoml.frameworks.h2o import H2oModelArtifact
    >>> from bentoml.adapters import DataframeInput
    >>>
    >>> @bentoml.artifacts([H2oModelArtifact('model')])
    >>> @bentoml.env(infer_pip_packages=True)
    >>> class H2oModelService(bentoml.BentoService):
    >>>
    >>>     @bentoml.api(input=DataframeInput(), batch=True)
    >>>     def predict(self, df):
    >>>         hf = h2o.H2OFrame(df)
    >>>         predictions = self.artifacts.model.predict(hf)
    >>>         return predictions.as_data_frame()
    >>>
    >>> svc = H2oModelService()
    >>>
    >>> svc.pack('model', model_to_save)
    """

    def __init__(self, name):
        super().__init__(name)

        self._model = None

    def set_dependencies(self, env: BentoServiceEnv):
        if env._infer_pip_packages:
            env.add_pip_packages(['h2o'])
        env.add_conda_dependencies(['openjdk'])

    def _model_file_path(self, base_path):
        return os.path.join(base_path, self.name)

    def pack(self, model, metadata=None):  # pylint:disable=arguments-differ
        self._model = model
        return self

    def load(self, path):
        try:
            import h2o
        except ImportError:
            raise MissingDependencyException(
                "h2o package is required to use H2oModelArtifact"
            )

        h2o.init()
        model = h2o.load_model(self._model_file_path(path))
        self._model = model
        return self

    def save(self, dst):
        try:
            import h2o
        except ImportError:
            raise MissingDependencyException(
                "h2o package is required to use H2oModelArtifact"
            )

        h2o_saved_path = h2o.save_model(model=self._model, path=dst, force=True)
        shutil.move(h2o_saved_path, self._model_file_path(dst))
        return

    def get(self):
        return self._model



import boto3

from bentoml.yatai.deployment.aws_ec2.constants import (
    BENTOSERVICE_PORT,
    AWS_EC2_IN_SERVICE_STATE,
    TARGET_HEALTHY_STATUS,
)
from bentoml.yatai.deployment.aws_utils import call_sam_command
from bentoml.exceptions import BentoMLException


def build_template(template_file_path, project_directory, region):
    status_code, stdout, stderr = call_sam_command(
        ["build", "-t", template_file_path], project_directory, region
    )

    if status_code != 0:
        error_message = stderr if stderr else stdout
        raise BentoMLException("Failed to build ec2 service {}".format(error_message))

    return status_code, stdout, stderr


def package_template(s3_bucket_name, project_directory, region):
    status_code, stdout, stderr = call_sam_command(
        [
            "package",
            "--output-template-file",
            "packaged.yaml",
            "--s3-bucket",
            s3_bucket_name,
        ],
        project_directory,
        region,
    )
    if status_code != 0:
        error_message = stderr if stderr else stdout
        raise BentoMLException("Failed to package ec2 service {}".format(error_message))
    return status_code, stdout, stderr


def deploy_template(stack_name, s3_bucket_name, project_directory, region):
    status_code, stdout, stderr = call_sam_command(
        [
            "deploy",
            "--template-file",
            "packaged.yaml",
            "--stack-name",
            stack_name,
            "--capabilities",
            "CAPABILITY_IAM",
            "--s3-bucket",
            s3_bucket_name,
        ],
        project_directory,
        region,
    )
    if status_code != 0:
        error_message = stderr if stderr else stdout
        raise BentoMLException("Failed to deploy ec2 service {}".format(error_message))
    return status_code, stdout, stderr


def get_endpoints_from_instance_address(instances, api_names):
    all_endpoints = []
    for instance in instances:
        if instance["state"] == AWS_EC2_IN_SERVICE_STATE:
            for api in api_names:
                all_endpoints.append(
                    "{ep}:{port}/{api}".format(
                        ep=instance["endpoint"], port=BENTOSERVICE_PORT, api=api
                    )
                )

    return all_endpoints


def get_healthy_target(target_group_arn, region):
    eb_client = boto3.client("elbv2", region)

    all_targets = eb_client.describe_target_health(TargetGroupArn=target_group_arn)
    for instance in all_targets["TargetHealthDescriptions"]:
        if instance["TargetHealth"]["State"] == TARGET_HEALTHY_STATUS:
            return instance
    return None



import os
from bentoml.exceptions import MissingDependencyException
from bentoml.service.artifacts import BentoServiceArtifact
from bentoml.service.env import BentoServiceEnv


class GluonModelArtifact(BentoServiceArtifact):
    """
    Abstraction for saving/loading gluon models

    Args:
        name (str): Name for the artifact

    Raises:
        MissingDependencyError: mxnet package is required for GluonModelArtifact

    Example usage:

    >>> from bentoml import env, artifacts, api, BentoService
    >>> from bentoml.adapters import JsonInput
    >>> from bentoml.frameworks.gluon import GluonModelArtifact
    >>> import mxnet as mx
    >>>
    >>> @env(infer_pip_packages=True)
    >>> @artifacts([GluonModelArtifact('model')])
    >>> class GluonClassifier(BentoService):
    >>>     @api(input=JsonInput(), batch=False)
    >>>     def predict(self, request):
    >>>         nd_input = mx.nd.array(request['input'])
    >>>     return self.artifacts.model(nd_input).asnumpy()
    >>>
    >>> svc = GluonClassifier()
    >>> svc.pack('model', model_to_save)
    >>> svc.save()
    """

    def __init__(self, name: str):
        super().__init__(name)
        self._model = None

    def pack(self, model, metadata: dict = None):  # pylint: disable=unused-argument
        try:
            import mxnet  # noqa # pylint: disable=unused-import
        except ImportError:
            raise MissingDependencyException(
                "mxnet package is required to use GluonModelArtifact"
            )
        self._model = model
        return self

    def load(self, path):
        try:
            from mxnet import gluon  # noqa # pylint: disable=unused-import
        except ImportError:
            raise MissingDependencyException(
                "mxnet package is required to use GluonModelArtifact"
            )

        prefix = self._model_file_path(path)
        model = gluon.nn.SymbolBlock.imports(
            "{}-symbol.json".format(prefix), ["data"], "{}-0000.params".format(prefix)
        )
        return self.pack(model)

    def _model_file_path(self, base_path):
        return os.path.join(base_path, self.name)

    def save(self, dst):
        self._model.export(self._model_file_path(dst))

    def get(self):
        return self._model

    def set_dependencies(self, env: BentoServiceEnv):
        if env._infer_pip_packages:
            env.add_pip_packages(["mxnet"])



import datetime
import enum

from sqlalchemy import UniqueConstraint, Column, Integer, Enum, DateTime, String
from bentoml.exceptions import LockUnavailable
from bentoml.yatai.db import Base


# LOCK_STATUS is an enum of the type of lock currently held
class LOCK_STATUS(enum.Enum):
    read_lock = 1
    write_lock = 2


class Lock(Base):
    __tablename__ = 'locks'
    __table_args__ = tuple(UniqueConstraint('resource_id', name='_resource_id_uc',))
    id = Column(Integer, primary_key=True)
    resource_id = Column(String, nullable=False, unique=True)
    lock_status = Column(Enum(LOCK_STATUS))
    locks_held = Column(Integer)
    ttl = Column(DateTime)

    # releases current lock
    def release(self, sess):
        self.locks_held -= 1

        # only delete lock row if no references left
        if self.locks_held == 0:
            sess.delete(self)

    # renews lock for `ttl_min` more min
    def renew(self, ttl_min):
        now = datetime.datetime.now()
        self.ttl = now + datetime.timedelta(minutes=ttl_min)


class LockStore(object):
    @staticmethod
    def _find_lock(sess, resource_id):
        lock_obj = sess.query(Lock).filter_by(resource_id=resource_id).first()
        return lock_obj

    # acquires a lock of type `lock_type` on `resource_id` that will expire in
    # `ttl_min` minutes
    @staticmethod
    def acquire(sess, lock_type, resource_id, ttl_min):
        now = datetime.datetime.now()
        ttl = now + datetime.timedelta(minutes=ttl_min)
        existing_lock = LockStore._find_lock(sess, resource_id)

        # no lock found; free to acquire
        if not existing_lock:
            # create lock with properties
            lock = Lock()
            lock.lock_status = lock_type
            lock.resource_id = resource_id
            lock.ttl = ttl
            lock.locks_held = 1
            sess.add(lock)
            return lock

        # existing lock expired, free to overwrite
        if existing_lock.ttl < now:
            existing_lock.lock_status = lock_type
            existing_lock.resource_id = resource_id
            existing_lock.ttl = ttl

            # locks_held can be held constant, we're just
            # overwriting existing lock
            return existing_lock

        # lock already exists
        if lock_type == LOCK_STATUS.read_lock:
            # acquire read lock
            if existing_lock.lock_status == LOCK_STATUS.write_lock:
                raise LockUnavailable("Failed to acquire read lock, write lock held")

            # current state is already read_lock, free to acquire read_lock
            # (multiple read_locks can be concurrently held)
            # bump ttl
            existing_lock.renew(ttl_min)
            existing_lock.locks_held += 1
            return existing_lock
        else:
            # acquire write lock
            # can't acquire write lock when any other lock is already held
            raise LockUnavailable("Failed to acquire write lock, another lock held")



import os
import json

from bentoml.exceptions import InvalidArgument, MissingDependencyException
from bentoml.service.artifacts import BentoServiceArtifact
from bentoml.service.env import BentoServiceEnv


class EasyOCRArtifact(BentoServiceArtifact):
    """
    Artifact class  for saving/loading EasyOCR models

    Args:
        name (str): Name for the artifact

    Raises:
        MissingDependencyError: easyocr>=1.3 package is required for EasyOCRArtifact
        InvalidArgument: invalid argument type, model being packed
            must be easyocr.easyocr.Reader

    Example usage:

    >>> import bentoml
    >>> from bentoml.frameworks.easyocr import EasyOCRArtifact
    >>> from bentoml.adapters import ImageInput
    >>>
    >>> @bentoml.env(pip_packages=["easyocr>=1.3.0"])
    >>> @bentoml.artifacts([EasyOCRArtifact("chinese_small")])
    >>> class EasyOCRService(bentoml.BentoService):
    >>>     @bentoml.api(input=ImageInput(), batch=False)
    >>>     def predict(self, image):
    >>>         reader = self.artifacts.chinese_small
    >>>         raw_results = reader.readtext(np.array(image))
    >>>         text_instances = [x[1] for x in raw_results]
    >>>         return {"text" : text_instances}
    >>>
    >>> import easyocr
    >>> service = EasyOCRService()
    >>>
    >>> lang_list = ['ch_sim', 'en']
    >>> recog_network = "zh_sim_g2"
    >>>
    >>> model = easyocr.Reader(lang_list=lang_list, download_enabled=True, \
    >>>                        recog_network=recog_network)
    >>> service.pack('chinese_small', model, lang_list=lang_list, \
    >>>              recog_network= recog_network)
    >>>
    >>> saved_path = service.save()

    """

    def __init__(self, name):
        super().__init__(name)

        self._model = None
        self._detect_model = None
        self._recog_network = None
        self._model_params = None
        self._gpu = None

    def set_dependencies(self, env: BentoServiceEnv):
        if env._infer_pip_packages:
            env.add_pip_packages(["easyocr>=1.3.0"])

    def _model_file_path(self, base_path):
        return os.path.join(base_path, self.name)

    def pack(
        self,
        easyocr_model,
        metadata=None,
        recog_network="english_g2",
        lang_list=None,
        detect_model="craft_mlt_25k",
        gpu=False,
    ):  # pylint:disable=arguments-differ
        try:
            import easyocr  # noqa # pylint: disable=unused-import

            assert easyocr.__version__ >= "1.3"
        except ImportError:
            raise MissingDependencyException(
                "easyocr>=1.3 package is required to use EasyOCRArtifact"
            )

        if not (type(easyocr_model) is easyocr.easyocr.Reader):
            raise InvalidArgument(
                "'easyocr_model' must be of type  easyocr.easyocr.Reader"
            )

        if not lang_list:
            lang_list = ['en']
        self._model = easyocr_model
        self._detect_model = detect_model
        self._recog_network = recog_network
        self._gpu = gpu
        self._model_params = {
            "lang_list": lang_list,
            "recog_network": recog_network,
            "gpu": gpu,
        }

        return self

    def load(self, path):
        try:
            import easyocr  # noqa # pylint: disable=unused-import

            assert easyocr.__version__ >= "1.3"
        except ImportError:
            raise MissingDependencyException(
                "easyocr>=1.3 package is required to use EasyOCRArtifact"
            )

        with open(os.path.join(path, f"{self._name}.json"), "r") as f:
            model_params = json.load(f)

        model = easyocr.Reader(
            model_storage_directory=path, download_enabled=False, **model_params
        )
        self._model = model

        return self.pack(model)

    def get(self):
        return self._model

    def save(self, dst):
        from shutil import copyfile

        src_folder = self._model.model_storage_directory

        detect_filename = f"{self._detect_model}.pth"

        if not os.path.exists(os.path.join(dst, detect_filename)):
            copyfile(
                os.path.join(src_folder, detect_filename),
                os.path.join(dst, detect_filename),
            )

        fname = self._recog_network + ".pth"
        copyfile(os.path.join(src_folder, fname), os.path.join(dst, fname))

        with open(os.path.join(dst, f"{self._name}.json"), "w") as f:
            json.dump(self._model_params, f)



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0


# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
import inspect
import logging
import multiprocessing
import os
import re
import subprocess
import sys
import tempfile
import threading
from typing import List
import uuid

from simple_di import Provide, inject

from bentoml.adapters import BaseInputAdapter, BaseOutputAdapter, DefaultOutput
from bentoml.configuration.containers import BentoMLContainer
from bentoml.exceptions import BentoMLException, InvalidArgument, NotFound
from bentoml.saved_bundle import save_to_dir
from bentoml.saved_bundle.config import (
    DEFAULT_MAX_BATCH_SIZE,
    DEFAULT_MAX_LATENCY,
    SavedBundleConfig,
)
from bentoml.saved_bundle.pip_pkg import seek_pip_packages
from bentoml.service.artifacts import ArtifactCollection, BentoServiceArtifact
from bentoml.service.env import BentoServiceEnv
from bentoml.service.inference_api import InferenceAPI
from bentoml.utils.hybridmethod import hybridmethod

ARTIFACTS_DIR_NAME = "artifacts"
BENTOML_RESERVED_API_NAMES = [
    "index",
    "swagger",
    "docs",
    "healthz",
    "metrics",
    "feedback",
]
logger = logging.getLogger(__name__)
prediction_logger = logging.getLogger("bentoml.prediction")


def validate_inference_api_name(api_name: str):
    if not api_name.isidentifier():
        raise InvalidArgument(
            "Invalid API name: '{}', a valid identifier may only contain letters,"
            " numbers, underscores and not starting with a number.".format(api_name)
        )

    if api_name in BENTOML_RESERVED_API_NAMES:
        raise InvalidArgument(
            "Reserved API name: '{}' is reserved for infra endpoints".format(api_name)
        )


def validate_inference_api_route(route: str):
    if re.findall(
        r"[?#]+|^(//)|^:", route
    ):  # contains '?' or '#' OR  start with '//' OR start with ':'
        # https://tools.ietf.org/html/rfc3986#page-22
        raise InvalidArgument(
            "The path {} contains illegal url characters".format(route)
        )
    if route in BENTOML_RESERVED_API_NAMES:
        raise InvalidArgument(
            "Reserved API route: '{}' is reserved for infra endpoints".format(route)
        )


def api_decorator(
    *args,
    input: BaseInputAdapter = None,
    output: BaseOutputAdapter = None,
    api_name: str = None,
    route: str = None,
    api_doc: str = None,
    mb_max_batch_size: int = DEFAULT_MAX_BATCH_SIZE,
    mb_max_latency: int = DEFAULT_MAX_LATENCY,
    batch=False,
    **kwargs,
):  # pylint: disable=redefined-builtin
    """
    A decorator exposed as `bentoml.api` for defining Inference API in a BentoService
    class.

    :param input: InputAdapter instance of the inference API
    :param output: OutputAdapter instance of the inference API
    :param api_name: API name, default to the user-defined callback function's function
        name
    :param route: Specify HTTP URL route of this inference API. By default,
        `api.name` is used as the route.  This parameter can be used for customizing
        the URL route, e.g. `route="/api/v2/model_a/predict"`
        Default: None (the same as api_name)
    :param api_doc: user-facing documentation of the inference API. default to the
        user-defined callback function's docstring
    :param mb_max_batch_size: The maximum size of requests batch accepted by this
        inference API. This parameter governs the throughput/latency trade off, and
        avoids having large batches that exceed some resource constraint (e.g. GPU
        memory to hold the entire batch's data). Default: 1000.
    :param mb_max_latency: The latency goal of this inference API in milliseconds.
        Default: 10000.


    Example usage:

    >>> from bentoml import BentoService, api
    >>> from bentoml.adapters import JsonInput, DataframeInput
    >>>
    >>> class FraudDetectionAndIdentityService(BentoService):
    >>>
    >>>     @api(input=JsonInput(), batch=True)
    >>>     def fraud_detect(self, json_list):
    >>>         # user-defined callback function that process inference requests
    >>>
    >>>     @api(input=DataframeInput(input_json_orient='records'), batch=True)
    >>>     def identity(self, df):
    >>>         # user-defined callback function that process inference requests
    """

    def decorator(func):
        _api_name = func.__name__ if api_name is None else api_name
        _api_route = _api_name if route is None else route
        validate_inference_api_name(_api_name)
        validate_inference_api_route(_api_route)
        _api_doc = func.__doc__ if api_doc is None else api_doc

        if input is None:
            # Raise error when input adapter class passed without instantiation
            if not args or not (
                inspect.isclass(args[0]) and issubclass(args[0], BaseInputAdapter)
            ):
                raise InvalidArgument(
                    "BentoService @api decorator first parameter must "
                    "be an instance of a class derived from "
                    "bentoml.adapters.BaseInputAdapter "
                )

            # noinspection PyPep8Naming
            InputAdapter = args[0]
            input_adapter = InputAdapter(*args[1:], **kwargs)
            output_adapter = DefaultOutput()
        else:
            assert isinstance(input, BaseInputAdapter), (
                "API input parameter must be an instance of a class derived from "
                "bentoml.adapters.BaseInputAdapter"
            )
            input_adapter = input
            output_adapter = output or DefaultOutput()

        setattr(func, "_is_api", True)
        setattr(func, "_input_adapter", input_adapter)
        setattr(func, "_output_adapter", output_adapter)
        setattr(func, "_api_name", _api_name)
        setattr(func, "_api_route", _api_route)
        setattr(func, "_api_doc", _api_doc)
        setattr(func, "_mb_max_batch_size", mb_max_batch_size)
        setattr(func, "_mb_max_latency", mb_max_latency)
        setattr(func, "_batch", batch)

        return func

    return decorator


def web_static_content_decorator(web_static_content):
    """Define web UI static files required to be bundled with a BentoService

    Args:
        web_static_content: path to directory containg index.html and static dir

    >>>  @web_static_content('./ui/')
    >>>  class MyMLService(BentoService):
    >>>     pass
    """

    def decorator(bento_service_cls):
        bento_service_cls._web_static_content = web_static_content
        return bento_service_cls

    return decorator


def artifacts_decorator(artifacts: List[BentoServiceArtifact]):
    """Define artifacts required to be bundled with a BentoService

    Args:
        artifacts (list(bentoml.artifact.BentoServiceArtifact)): A list of desired
            artifacts required by this BentoService
    """

    def decorator(bento_service_cls):
        artifact_names = set()
        for artifact in artifacts:
            if not isinstance(artifact, BentoServiceArtifact):
                raise InvalidArgument(
                    "BentoService @artifacts decorator only accept list of "
                    "BentoServiceArtifact instances, instead got type: '%s'"
                    % type(artifact)
                )

            if artifact.name in artifact_names:
                raise InvalidArgument(
                    "Duplicated artifact name `%s` detected. Each artifact within one"
                    "BentoService must have an unique name" % artifact.name
                )

            artifact_names.add(artifact.name)

        bento_service_cls._declared_artifacts = artifacts
        return bento_service_cls

    return decorator


def env_decorator(
    pip_dependencies: List[str] = None,
    pip_packages: List[str] = None,
    pip_index_url: str = None,
    pip_trusted_host: str = None,
    pip_extra_index_url: str = None,
    auto_pip_dependencies: bool = None,
    infer_pip_packages: bool = False,
    requirements_txt_file: str = None,
    conda_channels: List[str] = None,
    conda_overwrite_channels: bool = False,
    conda_override_channels: bool = False,
    conda_dependencies: List[str] = None,
    conda_env_yml_file: str = None,
    setup_sh: str = None,
    docker_base_image: str = None,
    zipimport_archives: List[str] = None,
):
    """Define environment and dependencies required for the BentoService being created

    Args:
        pip_packages:: list of pip_packages required, specified by package name
            or with specified version `{package_name}=={package_version}`
        pip_dependencies: same as pip_packages but deprecated
        pip_index_url: passing down to pip install --index-url option
        pip_trusted_host: passing down to pip install --trusted-host option
        pip_extra_index_url: passing down to pip install --extra-index-url option
        infer_pip_packages: whether to automatically find all the required
            pip dependencies and pin their version
        auto_pip_dependencies: same as infer_pip_packages but deprecated
        requirements_txt_file: path to the requirements.txt where pip dependencies
            are explicitly specified, with ideally pinned versions
        conda_channels: list of extra conda channels other than default channels to be
            used. This is equivalent to passing the --channels to conda commands
        conda_override_channels: ensures that conda searches only your specified
            channel and no other channels, such as default channels.
            This is equivalent to passing the --override-channels option to conda
            commands, or adding `nodefaults` to the `channels` in the environment.yml
        conda_overwrite_channels: aliases to `override_channels`
        conda_dependencies: list of conda dependencies required
        conda_env_yml_file: use a pre-defined conda environment yml file
        setup_sh: user defined setup bash script, it is executed in docker build time
        docker_base_image: used for customizing the docker container image built with
            BentoML saved bundle. Base image must either have both `bash` and `conda`
            installed; or have `bash`, `pip`, `python` installed, in which case the user
            is required to ensure the python version matches the BentoService bundle
        zipimport_archives: list of zipimport archives paths relative to the module path
    """

    if requirements_txt_file:
        if pip_packages:
            logger.warning("Ignoring pip_packages as requirements_txt_file is set.")
        if pip_index_url or pip_trusted_host or pip_extra_index_url:
            logger.warning(
                "Ignoring pip related options as requirements_txt_file is set."
            )
    if pip_dependencies is not None:
        logger.warning(
            "`pip_dependencies` parameter in `@env` is being deprecated soon, use "
            "`pip_packages` instead, e.g. `@env(pip_packages=[\"numpy\"])`"
        )
    if auto_pip_dependencies is not None:
        logger.warning(
            "`auto_pip_dependencies` parameter in `@env` is being deprecated soon,"
            "use `infer_pip_packages` instead, e.g. `@env(infer_pip_packages=True)`"
        )

    def decorator(bento_service_cls):
        bento_service_cls._env = BentoServiceEnv(
            pip_packages=pip_packages or pip_dependencies,
            pip_index_url=pip_index_url,
            pip_trusted_host=pip_trusted_host,
            pip_extra_index_url=pip_extra_index_url,
            infer_pip_packages=infer_pip_packages or auto_pip_dependencies,
            requirements_txt_file=requirements_txt_file,
            conda_channels=conda_channels,
            conda_override_channels=conda_override_channels,
            conda_overwrite_channels=conda_overwrite_channels,
            conda_dependencies=conda_dependencies,
            conda_env_yml_file=conda_env_yml_file,
            setup_sh=setup_sh,
            docker_base_image=docker_base_image,
            zipimport_archives=zipimport_archives,
        )
        return bento_service_cls

    return decorator


def ver_decorator(major, minor):
    """Decorator for specifying the version of a custom BentoService.

    Args:
        major (int): Major version number for Bento Service
        minor (int): Minor version number for Bento Service

    BentoML uses semantic versioning for BentoService distribution:

    * MAJOR is incremented when you make breaking API changes

    * MINOR is incremented when you add new functionality without breaking the
      existing API or functionality

    * PATCH is incremented when you make backwards-compatible bug fixes

    'Patch' is provided(or auto generated) when calling BentoService#save,
    while 'Major' and 'Minor' can be defined with '@ver' decorator

    >>>  from bentoml import ver, artifacts
    >>>  from bentoml.service.artifacts.common import PickleArtifact
    >>>
    >>>  @ver(major=1, minor=4)
    >>>  @artifacts([PickleArtifact('model')])
    >>>  class MyMLService(BentoService):
    >>>     pass
    >>>
    >>>  svc = MyMLService()
    >>>  svc.pack("model", trained_classifier)
    >>>  svc.set_version("2019-08.iteration20")
    >>>  svc.save()
    >>>  # The final produced BentoService bundle will have version:
    >>>  # "1.4.2019-08.iteration20"
    """

    def decorator(bento_service_cls):
        bento_service_cls._version_major = major
        bento_service_cls._version_minor = minor
        return bento_service_cls

    return decorator


def validate_version_str(version_str):
    """
    Validate that version str format is either a simple version string that:
        * Consist of only ALPHA / DIGIT / "-" / "." / "_"
        * Length between 1-128
    Or a valid semantic version https://github.com/semver/semver/blob/master/semver.md
    """
    regex = r"[A-Za-z0-9_.-]{1,128}\Z"
    semver_regex = r"^(?P<major>0|[1-9]\d*)\.(?P<minor>0|[1-9]\d*)\.(?P<patch>0|[1-9]\d*)(?:-(?P<prerelease>(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\+(?P<buildmetadata>[0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?$"  # noqa: E501
    if (
        re.match(regex, version_str) is None
        and re.match(semver_regex, version_str) is None
    ):
        raise InvalidArgument(
            'Invalid BentoService version: "{}", it can only consist'
            ' ALPHA / DIGIT / "-" / "." / "_", and must be less than'
            "128 characters".format(version_str)
        )

    if version_str.lower() == "latest":
        raise InvalidArgument('BentoService version can not be set to "latest"')


def save(bento_service, base_path=None, version=None, labels=None):
    """
    Save and register the given BentoService via BentoML's built-in model management
    system. BentoML by default keeps track of all the SavedBundle's files and metadata
    in local file system under the $BENTOML_HOME(~/bentoml) directory. Users can also
    configure BentoML to save their BentoService to a shared Database and cloud object
    storage such as AWS S3.

    :param bento_service: target BentoService instance to be saved
    :param base_path: optional - override repository base path
    :param version: optional - save with version override
    :param labels: optional - user defined labels

    :return: saved_path: file path to where the BentoService is saved
    """

    logger.warning(
        "`from bentoml import save` is being deprecated soon, use BentoService#save "
        "and BentoService#save_to_dir instead."
    )

    from bentoml.yatai.client import YataiClient
    from bentoml.yatai.yatai_service import get_yatai_service

    if base_path:
        yatai_service = get_yatai_service(file_system_directory=base_path)
        yatai_client = YataiClient(yatai_service)
    else:
        yatai_client = YataiClient()

    return yatai_client.repository.upload(bento_service, version, labels)


class BentoService:
    """
    BentoService is the base component for building prediction services using BentoML.

    BentoService provide an abstraction for describing model artifacts and environment
    dependencies required for a prediction service. And allows users to create inference
    APIs that defines the inferencing logic and how the underlying model can be served.
    Each BentoService can contain multiple models and serve multiple inference APIs.

    Usage example:

    >>>  from bentoml import BentoService, env, api, artifacts
    >>>  from bentoml.adapters import DataframeInput
    >>>  from bentoml.frameworks.sklearn import SklearnModelArtifact
    >>>
    >>>  @artifacts([SklearnModelArtifact('clf')])
    >>>  @env(pip_packages=["scikit-learn"])
    >>>  class MyMLService(BentoService):
    >>>
    >>>     @api(input=DataframeInput(), batch=True)
    >>>     def predict(self, df):
    >>>         return self.artifacts.clf.predict(df)
    >>>
    >>>  if __name__ == "__main__":
    >>>     bento_service = MyMLService()
    >>>     bento_service.pack('clf', trained_classifier_model)
    >>>     bento_service.save_to_dir('/bentoml_bundles')
    """

    # List of inference APIs that this BentoService provides
    _inference_apis: List[InferenceAPI] = []

    # Name of this BentoService. It is default the class name of this BentoService class
    _bento_service_name: str = None

    # For BentoService loaded from saved bundle, this will be set to the path of bundle.
    # When user install BentoService bundle as a PyPI package, this will be set to the
    # installed site-package location of current python environment
    _bento_service_bundle_path: str = None

    # List of artifacts required by this BentoService class, declared via the `@env`
    # decorator. This list is used for initializing an empty ArtifactCollection when
    # the BentoService class is instantiated
    _declared_artifacts: List[BentoServiceArtifact] = []

    # An instance of ArtifactCollection, containing all required trained model artifacts
    _artifacts: ArtifactCollection = None

    #  A `BentoServiceEnv` instance specifying the required dependencies and all system
    #  environment setups
    _env = None

    # When loading BentoService from saved bundle, this will be set to the version of
    # the saved BentoService bundle
    _bento_service_bundle_version = None

    # See `ver_decorator` function above for more information
    _version_major = None
    _version_minor = None

    # See `web_static_content` function above for more
    _web_static_content = None

    def __init__(self):
        # When creating BentoService instance from a saved bundle, set version to the
        # version specified in the saved bundle
        self._bento_service_version = self.__class__._bento_service_bundle_version
        self._dev_server_bundle_path: tempfile.TemporaryDirectory = None
        self._dev_server_interrupt_event: multiprocessing.Event = None
        self._dev_server_process: subprocess.Process = None

        self._config_artifacts()
        self._config_inference_apis()
        self._config_environments()

    def _config_environments(self):
        self._env = self.__class__._env or BentoServiceEnv()

        for api in self._inference_apis:
            self._env.add_pip_packages(api.input_adapter.pip_dependencies)
            self._env.add_pip_packages(api.output_adapter.pip_dependencies)

        for artifact in self.artifacts.get_artifact_list():
            artifact.set_dependencies(self.env)

    def _config_inference_apis(self):
        self._inference_apis = []

        for _, function in inspect.getmembers(
            self.__class__,
            predicate=lambda x: inspect.isfunction(x) or inspect.ismethod(x),
        ):
            if hasattr(function, "_is_api"):
                api_name = getattr(function, "_api_name")
                route = getattr(function, "_api_route", None)
                api_doc = getattr(function, "_api_doc")
                input_adapter = getattr(function, "_input_adapter")
                output_adapter = getattr(function, "_output_adapter")
                mb_max_latency = getattr(function, "_mb_max_latency")
                mb_max_batch_size = getattr(function, "_mb_max_batch_size")
                batch = getattr(function, "_batch")

                # Bind api method call with self(BentoService instance)
                user_func = function.__get__(self)

                self._inference_apis.append(
                    InferenceAPI(
                        self,
                        api_name,
                        api_doc,
                        input_adapter=input_adapter,
                        user_func=user_func,
                        output_adapter=output_adapter,
                        mb_max_latency=mb_max_latency,
                        mb_max_batch_size=mb_max_batch_size,
                        batch=batch,
                        route=route,
                    )
                )

    def _config_artifacts(self):
        self._artifacts = ArtifactCollection.from_declared_artifact_list(
            self._declared_artifacts
        )

        if self._bento_service_bundle_path:
            # For pip installed BentoService, artifacts directory is located at
            # 'package_path/artifacts/', but for loading from bundle directory, it is
            # in 'path/{service_name}/artifacts/'
            if os.path.isdir(
                os.path.join(self._bento_service_bundle_path, ARTIFACTS_DIR_NAME)
            ):
                artifacts_path = os.path.join(
                    self._bento_service_bundle_path, ARTIFACTS_DIR_NAME
                )
            else:
                artifacts_path = os.path.join(
                    self._bento_service_bundle_path, self.name, ARTIFACTS_DIR_NAME
                )

            self.artifacts.load_all(artifacts_path)

    @property
    def inference_apis(self) -> List[InferenceAPI]:
        """Return a list of user defined API functions

        Returns:
            list(InferenceAPI): List of Inference API objects
        """
        return self._inference_apis

    def get_inference_api(self, api_name):
        """Find the inference API in this BentoService with a specific name.

        When the api_name is None, this returns the first Inference API found in the
        `self.inference_apis` list.

        :param api_name: the target Inference API's name
        :return:
        """
        if api_name:
            try:
                return next(
                    (api for api in self.inference_apis if api.name == api_name)
                )
            except StopIteration:
                raise NotFound(
                    "Can't find API '{}' in service '{}'".format(api_name, self.name)
                )
        elif len(self.inference_apis) > 0:
            return self.inference_apis[0]
        else:
            raise NotFound(f"Can't find any inference API in service '{self.name}'")

    @property
    def artifacts(self):
        """ Returns the ArtifactCollection instance specified with this BentoService
        class

        Returns:
            artifacts(ArtifactCollection): A dictionary of packed artifacts from the
            artifact name to the BentoServiceArtifact instance
        """
        return self._artifacts

    @property
    def env(self):
        return self._env

    @property
    def web_static_content(self):
        return self._web_static_content

    def get_web_static_content_path(self):
        if not self.web_static_content:
            return None
        if self._bento_service_bundle_path:
            return os.path.join(
                self._bento_service_bundle_path, self.name, 'web_static_content',
            )
        else:
            return os.path.join(os.getcwd(), self.web_static_content)

    @hybridmethod
    @property
    def name(self):
        """
        :return: BentoService name
        """
        return self.__class__.name  # pylint: disable=no-value-for-parameter

    @name.classmethod
    @property
    def name(cls):  # pylint: disable=no-self-argument,invalid-overridden-method
        """
        :return: BentoService name
        """
        if cls._bento_service_name is not None:
            if not cls._bento_service_name.isidentifier():
                raise InvalidArgument(
                    'BentoService#_bento_service_name must be valid python identifier'
                    'matching regex `(letter|"_")(letter|digit|"_")*`'
                )

            return cls._bento_service_name
        else:
            # Use python class name as service name
            return cls.__name__

    def set_version(self, version_str=None):
        """Set the version of this BentoService instance. Once the version is set
        explicitly via `set_version`, the `self.versioneer` method will no longer be
        invoked when saving this BentoService.
        """
        if version_str is None:
            version_str = self.versioneer()

        if self._version_major is not None and self._version_minor is not None:
            # BentoML uses semantic versioning for BentoService distribution
            # when user specified the MAJOR and MINOR version number along with
            # the BentoService class definition with '@ver' decorator.
            # The parameter version(or auto generated version) here will be used as
            # PATCH field in the final version:
            version_str = ".".join(
                [str(self._version_major), str(self._version_minor), version_str]
            )

        validate_version_str(version_str)

        if self.__class__._bento_service_bundle_version is not None:
            logger.warning(
                "Overriding loaded BentoService(%s) version:%s to %s",
                self.__class__._bento_service_bundle_path,
                self.__class__._bento_service_bundle_version,
                version_str,
            )
            self.__class__._bento_service_bundle_version = None

        if (
            self._bento_service_version is not None
            and self._bento_service_version != version_str
        ):
            logger.warning(
                "Resetting BentoService '%s' version from %s to %s",
                self.name,
                self._bento_service_version,
                version_str,
            )

        self._bento_service_version = version_str
        return self._bento_service_version

    def versioneer(self):
        """
        Function used to generate a new version string when saving a new BentoService
        bundle. User can also override this function to get a customized version format
        """
        datetime_string = datetime.now().strftime("%Y%m%d%H%M%S")
        random_hash = uuid.uuid4().hex[:6].upper()

        # Example output: '20191009135240_D246ED'
        return datetime_string + "_" + random_hash

    @property
    def version(self):
        """
        Return the version of this BentoService. If the version of this BentoService has
        not been set explicitly via `self.set_version`, a new version will be generated
        with the `self.versioneer` method. User can customize this version str either by
        setting the version with `self.set_version` before a `save` call, or override
        the `self.versioneer` method to customize the version str generator logic.

        For BentoService loaded from a saved bundle, this will simply return the version
        information found in the saved bundle.

        :return: BentoService version str
        """
        if self.__class__._bento_service_bundle_version is not None:
            return self.__class__._bento_service_bundle_version

        if self._bento_service_version is None:
            self.set_version(self.versioneer())

        return self._bento_service_version

    @property
    def tag(self):
        """
        Bento tag is simply putting its name and version together, separated by a colon
        `tag` is mostly used in Yatai model management related APIs and operations
        """
        return f"{self.name}:{self.version}"

    def save(self, yatai_url=None, version=None, labels=None):
        """
        Save and register this BentoService via BentoML's built-in model management
        system. BentoML by default keeps track of all the SavedBundle's files and
        metadata in local file system under the $BENTOML_HOME(~/bentoml) directory.
        Users can also configure BentoML to save their BentoService to a shared Database
        and cloud object storage such as AWS S3.

        :param yatai_url: optional - URL path to Yatai server
        :param version: optional - save with version override
        :param labels: optional - labels dictionary
        :return: saved_path: file path to where the BentoService is saved
        """
        from bentoml.yatai.client import get_yatai_client

        yc = get_yatai_client(yatai_url)

        return yc.repository.upload(self, version, labels)

    def save_to_dir(self, path, version=None):
        """Save this BentoService along with all its artifacts, source code and
        dependencies to target file path, assuming path exist and empty. If target path
        is not empty, this call may override existing files in the given path.

        :param path (str): Destination of where the bento service will be saved
        :param version: optional - save with version override
        """
        return save_to_dir(self, path, version)

    @hybridmethod
    def pack(self, name, *args, **kwargs):
        """
        BentoService#pack method is used for packing trained model instances with a
        BentoService instance and make it ready for BentoService#save.

        pack(name, *args, **kwargs):

        :param name: name of the declared model artifact
        :param args: args passing to the target model artifact to be packed
        :param kwargs: kwargs passing to the target model artifact to be packed
        :return: this BentoService instance
        """
        self.artifacts.get(name).pack(*args, **kwargs)
        return self

    @pack.classmethod
    def pack(cls, *args, **kwargs):  # pylint: disable=no-self-argument
        """
        **Deprecated**: Legacy `BentoService#pack` class method, no longer supported
        """
        raise BentoMLException(
            "BentoService#pack class method is deprecated, use instance method `pack` "
            "instead. e.g.: svc = MyBentoService(); svc.pack('model', model_object)"
        )

    def get_bento_service_metadata_pb(self):
        return SavedBundleConfig(self).get_bento_service_metadata_pb()

    pip_dependencies_map = None

    def start_dev_server(self, port=None, enable_ngrok=False, debug=False):
        if self._dev_server_process:
            logger.warning(
                "There is already a running dev server, "
                "please call `service.stop_dev_server()` first."
            )
            return
        try:
            self._dev_server_bundle_path = tempfile.TemporaryDirectory()
            self.save_to_dir(self._dev_server_bundle_path.name)

            def print_log(p):
                for line in p.stdout:
                    print(line.decode(), end='')

            def run(path, interrupt_event):
                my_env = os.environ.copy()
                # my_env["FLASK_ENV"] = "development"
                cmd = [sys.executable, "-m", "bentoml", "serve"]
                if port:
                    cmd += ['--port', f'{port}']
                if enable_ngrok:
                    cmd += ['--run-with-ngrok']
                if debug:
                    cmd += ['--debug']
                cmd += [path]
                p = subprocess.Popen(
                    cmd,
                    stderr=subprocess.PIPE,
                    stdout=subprocess.PIPE,
                    stdin=subprocess.PIPE,
                    env=my_env,
                )
                threading.Thread(target=print_log, args=(p,), daemon=True).start()
                interrupt_event.wait()
                p.terminate()

            self._dev_server_interrupt_event = multiprocessing.Event()
            self._dev_server_process = multiprocessing.Process(
                target=run,
                args=(
                    self._dev_server_bundle_path.name,
                    self._dev_server_interrupt_event,
                ),
                daemon=True,
            )
            self._dev_server_process.start()
            logger.info(
                f"======= starting dev server on port: {port if port else 5000} ======="
            )
        except Exception as e:  # pylint: disable=broad-except
            self.stop_dev_server(skip_log=True)
            raise e

    def stop_dev_server(self, skip_log=False):
        if self._dev_server_interrupt_event:
            self._dev_server_interrupt_event.set()
            self._dev_server_interrupt_event = None
        if self._dev_server_process:
            self._dev_server_process.join()
            assert not self._dev_server_process.is_alive()
            self._dev_server_process = None
            logger.info("Dev server has stopped.")
        elif not skip_log:
            logger.warning("No dev server is running.")
        if self._dev_server_bundle_path:
            self._dev_server_bundle_path.cleanup()
            self._dev_server_bundle_path = None

    def __del__(self):
        if hasattr(self, "_dev_server_interrupt_event"):  # __init__ may not be called
            self.stop_dev_server(skip_log=True)

    @inject
    def infer_pip_dependencies_map(
        self,
        bentoml_version: str = Provide[
            BentoMLContainer.bento_bundle_deployment_version
        ],
    ):
        if not self.pip_dependencies_map:
            self.pip_dependencies_map = {}
            bento_service_module = sys.modules[self.__class__.__module__]
            if hasattr(bento_service_module, "__file__"):
                bento_service_py_file_path = bento_service_module.__file__
                reqs, unknown_modules = seek_pip_packages(bento_service_py_file_path)
                self.pip_dependencies_map.update(reqs)
                for module_name in unknown_modules:
                    logger.warning(
                        "unknown package dependency for module: %s", module_name
                    )

            # Reset bentoml to configured deploy version - this is for users using
            # customized BentoML branch for development but use a different stable
            # version for deployment
            #
            # For example, a BentoService created with local dirty branch will fail
            # to deploy with docker due to the version can't be found on PyPI, but
            # get_bentoml_deploy_version gives the user the latest released PyPI
            # version that's closest to the `dirty` branch
            self.pip_dependencies_map['bentoml'] = bentoml_version

        return self.pip_dependencies_map



# Copyright 2019 Atalaya Tech, Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os.path
import logging
import tempfile
import shutil
from bentoml.configuration import get_debug_mode


logger = logging.getLogger(__name__)


class TempDirectory(object):
    """
    Helper class that creates and cleans up a temporary directory, can
    be used as a context manager.

    >>> with TempDirectory() as tempdir:
    >>>     print(os.path.isdir(tempdir))
    """

    def __init__(
        self, cleanup=True, prefix="temp",
    ):

        self._cleanup = cleanup
        self._prefix = prefix
        self.path = None

    def __repr__(self):
        return "<{} {!r}>".format(self.__class__.__name__, self.path)

    def __enter__(self):
        self.create()
        return self.path

    def __exit__(self, exc_type, exc_val, exc_tb):
        if get_debug_mode():
            logger.debug(
                'BentoML in debug mode, keeping temp directory "%s"', self.path
            )
            return

        if self._cleanup:
            self.cleanup()

    def create(self):
        if self.path is not None:
            logger.debug("Skipped temp directory creation: %s", self.path)
            return self.path

        tempdir = tempfile.mkdtemp(prefix="bentoml-{}-".format(self._prefix))
        self.path = os.path.realpath(tempdir)
        logger.debug("Created temporary directory: %s", self.path)

    def cleanup(self, ignore_errors=False):
        """
        Remove the temporary directory created
        """
        if self.path is not None and os.path.exists(self.path):
            shutil.rmtree(self.path, ignore_errors=ignore_errors)
        self.path = None
